/*
 * File: $Id: ClinicalCase.java,v 1.1.1.1 $
 * History
 * Date: 12.12.2016  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2016 by
 *
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */
package org.spectrim.clinicalcasehandling;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import org.dcm4che2.data.DicomObject;
import org.spectrim.data.CSIData;
import org.spectrim.data.FileReaderWorker;
import org.spectrim.io.dicom.DicomHandler;
import org.spectrim.io.dicom.DicomHeader;

/**
 * The class handles clinical case related issues:
 * 
 * - Storage of most relevant patient information
 * - Storage of MRS source data information (Dicom format)
 * - Storage of fitting result data
 * 
 * - Note one case is not identical with a patient, but with one spectral 
 *   dataset which has been processed (either automatically, or manually). 
 * @author slotboom
 */
public class ClinicalCase implements ActionListener 
{
    /**
     * Creates a new clinical case: 
     * all initially required data is present in the DICOM header
     * file
     * @param dcm_spectro_file
     */
    public ClinicalCase( String dcm_spectro_file )
    {
       DicomObject spc_case = DicomHandler.getDicomObject( dcm_spectro_file );
        if ( spc_case != null )
        {
            DicomHeader header = DicomHandler.getHeader(spc_case);
            DicomHandler.readSiemensData(spc_case, header);
            
            String classUID = DicomHandler.getSOPClassUID(spc_case);
            String imagetype = DicomHandler.getImageType(spc_case);
            
            FileReaderWorker.filetype type;
            
            switch ( classUID ) 
            {    
                case "1.2.840.10008.5.1.4.1.1.4.2": // Spectro std
                    type = FileReaderWorker.filetype.MRS_STD;
                     m_isValidCase = true;
                    break;
                case "1.3.12.2.1107.5.9.1": // Spectro/Tensor Siemens
                    if (imagetype.indexOf("TENSOR") >= 0) 
                    {
                        type = FileReaderWorker.filetype.TENSOR;
                    } else 
                    {
                        type = FileReaderWorker.filetype.MRS_SIEMENS;
                         m_isValidCase = true;
                    }
                    break;
                case "1.2.840.10008.5.1.4.1.1.7": // Perfusion derived maps
                    type = FileReaderWorker.filetype.PERFUSION;
                    break;
                default: // Unknown format
                    type = FileReaderWorker.filetype.UNKNOWN;                                         
                    break;
            }
            
            // we are sure that this is a valid spectroscopy file
            if ( m_isValidCase == true )
            {
                m_case_UID = DicomHandler.getSOPInstanceUID(spc_case);
                m_patientName = DicomHandler.getPatientName(spc_case);
                m_scannerModel = DicomHandler.getModelName(spc_case);
                
                CSIData csidata;
                csidata = new CSIData( header, null );
                if ( ( csidata.nrVoxelsDir1 == 1 ) && 
                     ( csidata.nrVoxelsDir2 == 1 ) &&
                     ( csidata.nrVoxelsDir3 == 1 ) )
                {
                    m_spectroscopyDatatype = SVS_DATA;
                }
                else
                {
                    m_spectroscopyDatatype = CSI_DATA;
                }
                
            }

            m_studyData = DicomHandler.getStudyDate(spc_case);
            m_seriesNumber = DicomHandler.getSeriesNumber(spc_case) + "";
            m_seriesDescription = DicomHandler.getSeriesDescription(spc_case);
        }
        
    }
    
    
     @Override
    public void actionPerformed(ActionEvent e) 
    {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
    
    // Getters and Setters ..
    public String getM_spectroscopyFileReference() 
    {
        return m_spectroscopyFileReference;
    }

    public void setM_spectroscopyFileReference(String m_spectroscopyFileReference) 
    {
        this.m_spectroscopyFileReference = m_spectroscopyFileReference;
    }

    public int getM_spectroscopyDatatype() 
    {
        return m_spectroscopyDatatype;
    }

    public void setM_spectroscopyDatatype(int m_spectroscopyDatatype) 
    {
        this.m_spectroscopyDatatype = m_spectroscopyDatatype;
    }

    public String getM_patientName() {
        return m_patientName;
    }

    public void setM_patientName(String m_patientName) 
    {
        this.m_patientName = m_patientName;
    }

    public String getM_case_UID() 
    {
        return m_case_UID;
    }

    public void setM_case_UID(String m_case_UID) 
    {
        this.m_case_UID = m_case_UID;
    }

    public String getM_studyData() 
    {
        return m_studyData;
    }

    public void setM_studyData(String m_studyData) 
    {
        this.m_studyData = m_studyData;
    }

    public String getM_seriesDescription() 
    {
        return m_seriesDescription;
    }

    public void setM_seriesDescription(String m_seriesDescription) 
    {
        this.m_seriesDescription = m_seriesDescription;
    }

    public String getM_seriesNumber() 
    {
        return m_seriesNumber;
    }

    public void setM_seriesNumber(String m_seriesNumber) 
    {
        this.m_seriesNumber = m_seriesNumber;
    }

    public String getM_scannerModel() 
    {
        return m_scannerModel;
    }

    public void setM_scannerModel(String m_scannerModel) 
    {
        this.m_scannerModel = m_scannerModel;
    }

    public String getM_spectralDataType() 
    {
        return m_spectralDataType;
    }

    public void setM_spectralDataType(String m_spectralDataType) 
    {
        this.m_spectralDataType = m_spectralDataType;
    }

    public boolean isM_isValidCase() 
    {
        return m_isValidCase;
    }

    public void setM_isValidCase(boolean m_isValidCase) 
    {
        this.m_isValidCase = m_isValidCase;
    }
    
    public void setProcessingState( int processing_state )
    {
        m_processingState = processing_state;
    }
    
   // Attributes
    private String m_spectroscopyFileReference; 
    
    private int    m_spectroscopyDatatype;
    
    private String m_patientName;
    private String m_case_UID;          // chosen to be the same as the SOPInstanceUID
    private String m_studyData;
    
    private String m_seriesDescription;
    private String m_seriesNumber;
    
    private String m_scannerModel;
    
    private String m_spectralDataType;
    
    private int m_processingState;
    
    
    public final static int CSI_DATA = 1;
    public final static int SVS_DATA = 2;
    
    private boolean m_isValidCase = false;
    
    
    private int m_processingStatus;
    
    // currently 3 status values for processing state 
    public final static int DATA_NOT_PROCESSED      = 100;
    public final static int DATA_PROCESSING         = 101;
    public final static int DATA__PROCESSING_READY  = 102;

    
}
