package org.spectrim.fidsvd;

public class Fourier
{

//================================
//  Discrete Fourier Transform  
//================================

   public static double[] discreteFT(double[] data, int N, boolean forward) {
      double X[] = new double[2*N];
      double omega;
      int k, ki, kr, n;

//  If this is a inverse transform, reverse the
//  sign of the angle so the sin() terms will
//  change sign.

      if (forward) {
         omega = 2.0*Math.PI/N;
      } else {
         omega = -2.0*Math.PI/N;
      }

//  Perform the Discrete Fourier Transform.
//  The real and imaginary data are stored in the
//  x[] and X[] vectors as pairs, one real and
//  one imaginary value for each sample point.

      for(k=0; k<N; ++k) {
         kr = 2*k;
         ki = 2*k + 1;
         X[kr] = 0.0;
         X[ki] = 0.0;
         for(n=0; n<N; ++n) {
           X[kr] += data[2*n]*Math.cos(omega*n*k) + data[2*n+1]*Math.sin(omega*n*k);
           X[ki] += -data[2*n]*Math.sin(omega*n*k) + data[2*n+1]*Math.cos(omega*n*k);
         }
      }

//  If this is a forward transform, multiply
//  in the 1/N terms

      if ( forward ) {
         for(k=0; k<N; ++k) {
            X[2*k] /= N;
            X[2*k + 1] /= N;
         }
      }

//  Return the transformed data.

      return X;
   }

//================================
//  Fast Fourier Transform  
//
//================================

   public static void fastFT(double[] data, int N, boolean forward) {
      double omega, tempr, tempi, scale;
      double xtemp, cos, sin, xr, xi;
      int i, j, k, n, m, M;

//  bit-shift the input data. Remember that N
//  is the number of samples. There is a real and
//  a complex component to each data point.

      j=0;
      for(i=0; i<N-1; i++) {
         if (i<j) {
            tempr = data[2*i];
            tempi = data[2*i + 1];
            data[2*i] = data[2*j];
            data[2*i + 1] = data[2*j + 1];
            data[2*j] = tempr;
            data[2*j + 1] = tempi;
         }
         k = N/2;
         while (k <= j)  {
            j -= k;
            k >>= 1;
         }
         j += k;
      }

//  Perform the FFT. Recursively build up the
//  full transform from the sub-transforms

      if (forward) {
         scale = 1.0;
      } else {
         scale = -1.0;
      }

      M = 2;
      while( M < 2*N ) {
         omega = scale*2.0*Math.PI/M;
         sin = Math.sin(omega);
         cos = Math.cos(omega) - 1.0;
         xr = 1.0;
         xi = 0.0;
         for (m=0; m<M-1; m+=2) {
            for (i=m; i<2*N; i+=M*2) {
               j = i + M;
               tempr = xr*data[j] - xi*data[j+1];
               tempi = xr*data[j+1] + xi*data[j];
               data[j] = data[i] - tempr;
               data[j+1] = data[i+1] - tempi;
               data[i] += tempr;
               data[i+1] += tempi;
            }
            xtemp = xr;
            xr = xr + xr*cos - xi*sin;
            xi = xi + xtemp*sin + xi*cos;
         }
         M *= 2;
      }

//  If this is a forward transform, multiply
//  in the 1/N terms

      if ( forward ) {
         for(k=0; k<N; ++k) {
            data[2*k] /= N;
            data[2*k + 1] /= N;
         }
      }

      return;
   }
   
//================================
//  Fast Fourier Transform  
//  Same as the method above, only
//  with returning data to caller   
//================================

   public static double[] fastFT_(double[] data, int N, boolean forward) {
      double omega, tempr, tempi, scale;
      double xtemp, cos, sin, xr, xi;
      int i, j, k, n, m, M;

//  bit-shift the input data. Remember that N
//  is the number of samples. There is a real and
//  a complex component to each data point.

      j=0;
      for(i=0; i<N-1; i++) {
         if (i<j) {
            tempr = data[2*i];
            tempi = data[2*i + 1];
            data[2*i] = data[2*j];
            data[2*i + 1] = data[2*j + 1];
            data[2*j] = tempr;
            data[2*j + 1] = tempi;
         }
         k = N/2;
         while (k <= j)  {
            j -= k;
            k >>= 1;
         }
         j += k;
      }

//  Perform the FFT. Recursively build up the
//  full transform from the sub-transforms

      if (forward) {
         scale = 1.0;
      } else {
         scale = -1.0;
      }

      M = 2;
      while( M < 2*N ) {
         omega = scale*2.0*Math.PI/M;
         sin = Math.sin(omega);
         cos = Math.cos(omega) - 1.0;
         xr = 1.0;
         xi = 0.0;
         for (m=0; m<M-1; m+=2) {
            for (i=m; i<2*N; i+=M*2) {
               j = i + M;
               tempr = xr*data[j] - xi*data[j+1];
               tempi = xr*data[j+1] + xi*data[j];
               data[j] = data[i] - tempr;
               data[j+1] = data[i+1] - tempi;
               data[i] += tempr;
               data[i+1] += tempi;
            }
            xtemp = xr;
            xr = xr + xr*cos - xi*sin;
            xi = xi + xtemp*sin + xi*cos;
         }
         M *= 2;
      }

//  If this is a forward transform, multiply
//  in the 1/N terms

      if ( forward ) {
         for(k=0; k<N; ++k) {
            data[2*k] /= N;
            data[2*k + 1] /= N;
         }
      }

      return data;
   }
   
   
   
}