/*
 * File: $Id: CSIGrid.java,v 1.1.1.1 2006/04/19 09:38:47 hans Exp $
 * History
 * Date: 18.03.2005  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2003 - 2005 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


/**
 *
 *  A spectroscopic CSI voxel looks as follows is a cube
 *
 *
 *               x6  /------------------------/ x7 
 *                 / |                      / |  
 *               /   |                    /   |
 *             /     |                  /     |
 *           /       |                /       |                x                  
 *       x2 ------------------------- x3      |                |
 *          |        |               |        |                |
 *          |        |       Origin  |        |                /------ z
 *          |        |      O        |        |               /
 *          |     x5 / ---------------------- / x8           y
 *          |       /                |      /
 *          |     /                  |     /       
 *          |   /                    |   /          
 *          | /                      | /
 *          ------------------------- x4
 *         x1
 *
 *
 *  For a transverse Image  Gx, Gy, Gz
 *  For a Sagital    Image  Gz, Gy, Gx
 *  For a Coronal    Image  Gx, Gz, Gy
 *
 * If you have a Sagital description, voxel points ( x, y, z ) are given by:
 * 
 * ( sSpecPara.sVoI.dThickness, sSpecPara.sVoI.dPhaseFOV, sSpecPara.sVoI.dReadoutFOV )
 *
 * If you have a Transversal description, voxel points ( x, y, z ) of the 
 * voxel are given by:
 *
 * ( sSpecPara.sVoI.dReadoutFOV, sSpecPara.sVoI.dPhaseFOV, sSpecPara.sVoI.dThickness )
 *
 * If you have a Coronal description voxel points ( x, y, z ) of the  
 * voxel are given  by:
 *
 * ( sSpecPara.sVoI.dPhaseFOV, sSpecPara.sVoI.dThickness, sSpecPara.sVoI.dReadoutFOV )
 * @author  hans
 */


/*
 * CSIGrid.java
 *
 * Created on March 18, 2005, 1:47 PM
 */

package org.spectrim.imagegraphicobjects;

import java.awt.Point;
import java.awt.Graphics;
import java.awt.Color;

import java.util.LinkedList;
import java.util.Iterator;

/**
 * Class that defines a CSI grid 
 * @author  hans
 */
public class CSIGrid 
{
    
    /** 
     * Creates a new instance of CSIGrid. 
     * @param centerPoint
     * @param gridDirectionUnitVector1
     * @param gridDirectionUnitVector2
     * @param gridDirectionUnitVector3
     * @param gridSizeDir1
     * @param gridSizeDir2
     * @param gridSizeDir3
     * @param FOVDir1
     * @param FOVDir2
     * @param FOVDir3
     * @param selectedVoxelDir1
     * @param selectedVoxelDir2
     * @param selectedVoxelDir3
     */
    public CSIGrid( Point3D centerPoint, Point3D gridDirectionUnitVector1,
                    Point3D gridDirectionUnitVector2, Point3D gridDirectionUnitVector3,
                    int gridSizeDir1, int gridSizeDir2, int gridSizeDir3, 
                    double FOVDir1, double FOVDir2, double FOVDir3,
                    int selectedVoxelDir1, int selectedVoxelDir2, int selectedVoxelDir3 ) 
    {
        
        m_centerPoint = centerPoint;
        
        m_gridSizeDir1 = gridSizeDir1;
        m_gridSizeDir2 = gridSizeDir2;
        m_gridSizeDir3 = gridSizeDir3;
        
        m_FOVDir1 = FOVDir1;
        m_FOVDir2 = FOVDir2;
        m_FOVDir3 = FOVDir3;
        
        m_gridDirectionUnitVector1 = gridDirectionUnitVector1;
        m_gridDirectionUnitVector2 = gridDirectionUnitVector2;
        m_gridDirectionUnitVector3 = gridDirectionUnitVector3;
     
        m_selectedVoxelDir1 = selectedVoxelDir1;
        m_selectedVoxelDir2 = selectedVoxelDir2;
        m_selectedVoxelDir3 = selectedVoxelDir3;   
    }
    
    
    /**
     * Performs a translation of the voxel about translation.
     * @param translation
     */
    public void translateVoxel( Point3D translation )
    {
        m_currentVoxel_c1 = m_currentVoxel_c1.translate( translation );
        m_currentVoxel_c2 = m_currentVoxel_c2.translate( translation );
        m_currentVoxel_c3 = m_currentVoxel_c3.translate( translation );
        m_currentVoxel_c4 = m_currentVoxel_c4.translate( translation );
        m_currentVoxel_c5 = m_currentVoxel_c5.translate( translation );
        m_currentVoxel_c6 = m_currentVoxel_c6.translate( translation );
        m_currentVoxel_c7 = m_currentVoxel_c7.translate( translation );
        m_currentVoxel_c8 = m_currentVoxel_c8.translate( translation );
    }
    
    
    /**
     * Paints it self on a Graphics context g on a plane that has 
     *               p  as topmost point
     *               r1 as direction vector1    with FOV = FOVX
     *               r2 aS direction vector2    with FOV = FOVY
     */
    public void paintComponent( Graphics g, Point3D p, Point3D r1, Point3D r2, 
                                double FOVX, double FOVY, int NpixX, int NpixY ,
                                String comment, String title_comment )
    {
        paintCSIGrid( g, p, r1, r2, FOVX, FOVY, 
                      NpixX, NpixY , comment, title_comment );
    }
    
    
    /**
     * Paints the CSI grid based on constructor information
     */
    public void paintCSIGrid( Graphics g, Point3D p, Point3D r1, Point3D r2, 
                              double FOVX, double FOVY, int NpixX, int NpixY ,
                              String comment, String title_comment )
    {
        
        m_g = g;
        m_p = p;
        m_r1 = r1;
        m_r2 = r2;
        m_FOVX = FOVX;
        m_FOVY = FOVY;
        m_NpixX = NpixX;
        m_NpixY = NpixY;
        m_comment = comment;
        m_title_comment = title_comment;
        
        m_gridPointXY = new LinkedList<Point>( );
        
        for ( int k = 0 ; k < this.m_gridSizeDir3; k++ )
        {
            for ( int j = 0 ; j < this.m_gridSizeDir2; j++ )
            {
                for ( int i = 0 ; i < this.m_gridSizeDir1; i++ )
                {
                    
                    Point pnt = paintCSIVoxel( i, j, k, g, p, r1, r2, FOVX, FOVY, 
                                               NpixX, NpixY, comment, title_comment, Color.GREEN );
                    m_gridPointXY.add( pnt );
                }
            }
        }
        
        paintCSIVoxel( m_selectedVoxelDir1 - 1, m_selectedVoxelDir2 - 1, m_selectedVoxelDir3, 
                       g, p, r1, r2, FOVX, FOVY, NpixX, NpixY, comment, 
                       title_comment, Color.BLUE );
    }
    
    
    /**
     * Paints the CSI grid based on constructor information
     */
    private void paintCSIGrid(  )
    {

        m_gridPointXY = new LinkedList<Point>( );
        
        for ( int k = 0 ; k < this.m_gridSizeDir3; k++ )
        {
            for ( int j = 0 ; j < this.m_gridSizeDir2; j++ )
            {
                for ( int i = 0 ; i < this.m_gridSizeDir1; i++ )
                {
                    
                    Point pnt = paintCSIVoxel( i, j, k, m_g, m_p, m_r1, m_r2, m_FOVX, m_FOVY, 
                                               m_NpixX, m_NpixY, m_comment, m_title_comment, Color.GREEN );
                    m_gridPointXY.add( pnt );
                }
            }
        }
        
        paintCSIVoxel( m_selectedVoxelDir1 - 1, m_selectedVoxelDir2 - 1, m_selectedVoxelDir3, 
                       m_g, m_p, m_r1, m_r2, m_FOVX, m_FOVY, m_NpixX, m_NpixY, m_comment, 
                       m_title_comment, Color.BLUE );
    }
    
    
    /**
     * Sets the selected voxel
     */
    public void setSelectedVoxel( int dim1Point, int dim2Point, int dim3Point )
    {
        m_selectedVoxelDir1 = dim1Point;
        m_selectedVoxelDir2 = dim2Point;
        m_selectedVoxelDir3 = dim3Point;
        
        repaint( );
    }
    
    /**
     * Computes the 8 coordinates of a basic voxel which is determined by the 
     * FOV[X,Y,Z] and the number of gridpoints
     */
    private void computeBasicVoxel( )
    { 
        
        m_currentVoxel_c1 = new Point3D( 0 , 0 , 0 );
        
        m_currentVoxel_c2 = m_gridDirectionUnitVector2.scalarMultiplication( m_FOVDir2 / (double)m_gridSizeDir2 );
        m_currentVoxel_c4 = m_gridDirectionUnitVector1.scalarMultiplication( m_FOVDir1 / (double)m_gridSizeDir1 );
        m_currentVoxel_c5 = m_gridDirectionUnitVector3.scalarMultiplication( m_FOVDir3 / (double)m_gridSizeDir3 );
        
        m_currentVoxel_c3.m_x1 = m_currentVoxel_c2.m_x1 + m_currentVoxel_c4.m_x1;
        m_currentVoxel_c3.m_x2 = m_currentVoxel_c2.m_x2 + m_currentVoxel_c4.m_x2;
        m_currentVoxel_c3.m_x3 = m_currentVoxel_c2.m_x3 + m_currentVoxel_c4.m_x3;
        
        m_currentVoxel_c6.m_x1 = m_currentVoxel_c2.m_x1 + m_currentVoxel_c5.m_x1;
        m_currentVoxel_c6.m_x2 = m_currentVoxel_c2.m_x2 + m_currentVoxel_c5.m_x2;
        m_currentVoxel_c6.m_x3 = m_currentVoxel_c2.m_x3 + m_currentVoxel_c5.m_x3;

        m_currentVoxel_c8.m_x1 = m_currentVoxel_c4.m_x1 + m_currentVoxel_c5.m_x1;
        m_currentVoxel_c8.m_x2 = m_currentVoxel_c4.m_x2 + m_currentVoxel_c5.m_x2;
        m_currentVoxel_c8.m_x3 = m_currentVoxel_c4.m_x3 + m_currentVoxel_c5.m_x3;
        
        m_currentVoxel_c7.m_x1 = m_currentVoxel_c2.m_x1 + m_currentVoxel_c4.m_x1 + m_currentVoxel_c5.m_x1;
        m_currentVoxel_c7.m_x2 = m_currentVoxel_c2.m_x2 + m_currentVoxel_c4.m_x2 + m_currentVoxel_c5.m_x2;
        m_currentVoxel_c7.m_x3 = m_currentVoxel_c2.m_x3 + m_currentVoxel_c4.m_x3 + m_currentVoxel_c5.m_x3;
        
    }
    
    /**
     * Translates the basic voxel to a position belonging to their coordinates
     */ 
    public void computeTranslatedBasicVoxel( int i, int j, int k )
    {
        double shiftDir1 = ( i - ( m_gridSizeDir1 / 2.0 ) ) * this.m_FOVDir1 / m_gridSizeDir1;
        double shiftDir2 = ( j - ( m_gridSizeDir2 / 2.0 ) ) * this.m_FOVDir2 / m_gridSizeDir2;
        double shiftDir3 = ( k - ( m_gridSizeDir3 / 2.0 ) ) * this.m_FOVDir3 / m_gridSizeDir3;
        
        Point3D m_translate1 = new Point3D( );
        Point3D m_translate2 = new Point3D( );
        Point3D m_translate3 = new Point3D( );
        
        m_translate1 = this.m_gridDirectionUnitVector1.scalarMultiplication( shiftDir1 );
        m_translate2 = this.m_gridDirectionUnitVector2.scalarMultiplication( shiftDir2 );
        m_translate3 = this.m_gridDirectionUnitVector3.scalarMultiplication( shiftDir3 );
        
        Point3D translate = new Point3D( );
        
        translate.m_x1 = m_translate1.m_x1 + m_translate2.m_x1 + m_translate3.m_x1;
        translate.m_x2 = m_translate1.m_x2 + m_translate2.m_x2 + m_translate3.m_x2;
        translate.m_x3 = m_translate1.m_x3 + m_translate2.m_x3 + m_translate3.m_x3;
        
        m_currentVoxel_c1 = m_currentVoxel_c1.translate( translate );
        m_currentVoxel_c2 = m_currentVoxel_c2.translate( translate );
        m_currentVoxel_c3 = m_currentVoxel_c3.translate( translate );
        m_currentVoxel_c4 = m_currentVoxel_c4.translate( translate );
        m_currentVoxel_c5 = m_currentVoxel_c5.translate( translate );
        m_currentVoxel_c6 = m_currentVoxel_c6.translate( translate );
        m_currentVoxel_c7 = m_currentVoxel_c7.translate( translate );
        m_currentVoxel_c8 = m_currentVoxel_c8.translate( translate ); 
    }
    
    
    /**
     * This translation is due to the translation of the excited voxel, which
     * is fixed to the csi grid (at least in the Siemens implementation
     */ 
    private void computeGridTranslation( )
    {
        m_currentVoxel_c1 = m_currentVoxel_c1.translate( m_centerPoint );
        m_currentVoxel_c2 = m_currentVoxel_c2.translate( m_centerPoint );
        m_currentVoxel_c3 = m_currentVoxel_c3.translate( m_centerPoint );
        m_currentVoxel_c4 = m_currentVoxel_c4.translate( m_centerPoint );
        m_currentVoxel_c5 = m_currentVoxel_c5.translate( m_centerPoint );
        m_currentVoxel_c6 = m_currentVoxel_c6.translate( m_centerPoint );
        m_currentVoxel_c7 = m_currentVoxel_c7.translate( m_centerPoint );
        m_currentVoxel_c8 = m_currentVoxel_c8.translate( m_centerPoint ); 
    }
    
    
    /**
     * Paints it self on a Graphics context g on a plane that has 
     *               p  as topmost point
     *               r1 as direction vector1    with FOV = FOVX
     *               r2 aS direction vector2    with FOV = FOVY
     */
    public Point paintCSIVoxel( int i, int j, int k ,
                               Graphics g, Point3D p, Point3D r1, Point3D r2, 
                               double FOVX, double FOVY, int NpixX, int NpixY ,
                               String comment, String title_comment, Color color )
    {
        
        Point3D dummy = new Point3D( );
        
        computeBasicVoxel( );
        
        computeTranslatedBasicVoxel( i, j, k );
        
        computeGridTranslation( );
        
        // now we have computed the 8 edge points of the voxel
        computePlanesThatEncapsulateCurrentVoxel( );
        
        
        Point3D pc1 = dummy.projectPointOnPlane( m_currentVoxel_c1, p, r1, r2 );
        Point3D pc2 = dummy.projectPointOnPlane( m_currentVoxel_c2, p, r1, r2 );
        Point3D pc3 = dummy.projectPointOnPlane( m_currentVoxel_c3, p, r1, r2 );
        Point3D pc4 = dummy.projectPointOnPlane( m_currentVoxel_c4, p, r1, r2 );
        Point3D pc5 = dummy.projectPointOnPlane( m_currentVoxel_c5, p, r1, r2 );
        Point3D pc6 = dummy.projectPointOnPlane( m_currentVoxel_c6, p, r1, r2 );
        Point3D pc7 = dummy.projectPointOnPlane( m_currentVoxel_c7, p, r1, r2 );
        Point3D pc8 = dummy.projectPointOnPlane( m_currentVoxel_c8, p, r1, r2 );
        
        Point p1 = calculate2DPoint( pc1, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p2 = calculate2DPoint( pc2, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p3 = calculate2DPoint( pc3, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p4 = calculate2DPoint( pc4, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p5 = calculate2DPoint( pc5, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p6 = calculate2DPoint( pc6, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p7 = calculate2DPoint( pc7, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p8 = calculate2DPoint( pc8, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        
        int averageX = ( p1.x + p2.x + p3.x + p4.x + p5.x + p6.x + p7.x + p8.x ) / 8;
        int averageY = ( p1.y + p2.y + p3.y + p4.y + p5.y + p6.y + p7.y + p8.y ) / 8;
        
        g.setColor( color );
        g.drawLine( p1.x, p1.y, p2.x, p2.y );
        g.setColor( color );
        g.drawLine( p2.x, p2.y, p3.x, p3.y );
        g.setColor( color );
        g.drawLine( p3.x, p3.y, p4.x, p4.y );
        g.setColor( color );
        g.drawLine( p4.x, p4.y, p1.x, p1.y );
        
        g.setColor( color );
        g.drawLine( p1.x, p1.y, p5.x, p5.y );
        g.drawLine( p2.x, p2.y, p6.x, p6.y );
        g.drawLine( p3.x, p3.y, p7.x, p7.y );
        g.drawLine( p4.x, p4.y, p8.x, p8.y );
        
        g.setColor( color );
        g.drawLine( p5.x, p5.y, p6.x, p6.y );
        g.setColor( color );
        g.drawLine( p6.x, p6.y, p7.x, p7.y );
        g.setColor( color );
        g.drawLine( p7.x, p7.y, p8.x, p8.y );
        g.setColor( color );
        g.drawLine( p8.x, p8.y, p5.x, p5.y );
        
        if ( comment != null )
        {
            if ( comment.length( ) > 0 )
            {
                g.setColor( Color.green );
                g.drawString( comment, 5, NpixY -5 );
            }
        }
        
        if ( title_comment != null )
        {
            if ( title_comment.length( ) > 0 )
            {
                g.setColor( Color.green );
                g.drawString( title_comment, 5, 15 );
            }
        }
        
        return new Point( averageX, averageY );
    }
    
    /**
     * Computes the 2D point from 3D point in a plane that has
     *
     *               p  as topmost point
     *               r1 as direction vector1    with FOV = FOVX
     *               r2 aS direction vector2    with FOV = FOVY
     *               NpixX number of points in the r1 direction
     *               NpixY number of points in the r2 direction
     *
     */
    private Point calculate2DPoint( Point3D a, Point3D p, Point3D r1, Point3D r2,
                                    double FOVX, double FOVY, int NpixX, int NpixY )
    {
        
        double solLambda = 0;
        double solMhu = 0;
        
        double deltaX = a.m_x1 - p.m_x1;
        double deltaY = a.m_x2 - p.m_x2;
        double deltaZ = a.m_x3 - p.m_x3;
        
        // try to find solution from x, y equation
        double numeratorXY_mhu = r1.m_x2 * deltaX - r1.m_x1 * deltaY;
        double dominatorXY_mhu = r1.m_x2 * r2.m_x1 - r1.m_x1 * r2.m_x2;
       
        double numeratorXY_lambda = deltaX * r2.m_x2 - deltaY * r2.m_x1;
        double dominatorXY_lambda = r1.m_x1 * r2.m_x2 - r1.m_x2 * r2.m_x1;
        
        
        // try to find solution from x, z equation
        double numeratorXZ_mhu = r1.m_x3 * deltaX - r1.m_x1 * deltaZ;
        double dominatorXZ_mhu = r1.m_x3 * r2.m_x1 - r1.m_x1 * r2.m_x3;
        
        double numeratorXZ_lambda = deltaX * r2.m_x3 - deltaZ * r2.m_x1;
        double dominatorXZ_lambda = r1.m_x1 * r2.m_x3 - r1.m_x3 * r2.m_x1;
        
        
        // try to find solution from y, z equation
        double numeratorYZ_mhu = r1.m_x3 * deltaY - r1.m_x2 * deltaZ;
        double dominatorYZ_mhu = r1.m_x3 * r2.m_x2 - r1.m_x2 * r2.m_x3;
        
        double numeratorYZ_lambda = deltaY * r2.m_x3 - deltaZ * r2.m_x2;
        double dominatorYZ_lambda = r1.m_x2 * r2.m_x3 - r1.m_x3 * r2.m_x2;
        
        if ( ( Math.abs( dominatorXY_mhu ) > 1.0e-9 ) && ( Math.abs( dominatorXY_lambda ) > 1.0e-9 ) )
        {
            solLambda = ( numeratorXY_lambda / dominatorXY_lambda );
            solMhu    = ( numeratorXY_mhu / dominatorXY_mhu );
        }
         
        if ( ( Math.abs( dominatorXZ_mhu ) > 1.0e-9 ) && ( Math.abs( dominatorXZ_lambda ) > 1.0e-9 ) )
        {
            solLambda = ( numeratorXZ_lambda / dominatorXZ_lambda );
            solMhu    = ( numeratorXZ_mhu / dominatorXZ_mhu );
        }

        if ( ( Math.abs( dominatorYZ_mhu ) > 1.0e-9 ) && ( Math.abs( dominatorYZ_lambda ) > 1.0e-9 ) )
        {
            solLambda = ( numeratorYZ_lambda / dominatorYZ_lambda );
            solMhu    = ( numeratorYZ_mhu / dominatorYZ_mhu );
        }

        Point P = new Point( );
        P.x = (int)( (float)( solLambda * NpixX ) / FOVX );
        P.y = (int)( (float)( solMhu * NpixY ) / FOVY );
        
        return P;
    }
    
    /**
     * Computes the frontplane; backplane, leftplane, rightplane and
     * top and bottom plane
     */
    private void computePlanesThatEncapsulateCurrentVoxel( )
    {
        // Calculate the 6 border planes of the voxel
        
        // Plane 1: front plane through x1, x2, x4
        Point3D frontPlane_r1 = new Point3D( m_currentVoxel_c1.m_x1 - m_currentVoxel_c2.m_x1, 
                                             m_currentVoxel_c1.m_x2 - m_currentVoxel_c2.m_x2, 
                                             m_currentVoxel_c1.m_x3 - m_currentVoxel_c2.m_x3 );
        
        Point3D frontPlane_r2 = new Point3D( m_currentVoxel_c1.m_x1 - m_currentVoxel_c4.m_x1, 
                                             m_currentVoxel_c1.m_x2 - m_currentVoxel_c4.m_x2, 
                                             m_currentVoxel_c1.m_x3 - m_currentVoxel_c4.m_x3 );
        
        m_currentFrontVoxelPlane = new Plane3D( m_currentVoxel_c1, frontPlane_r1, frontPlane_r2 );
        
        // Plane 2: back plane through x5, x6, x8
        Point3D backPlane_r1 = new Point3D( m_currentVoxel_c5.m_x1 - m_currentVoxel_c6.m_x1, 
                                            m_currentVoxel_c5.m_x2 - m_currentVoxel_c6.m_x2, 
                                            m_currentVoxel_c5.m_x3 - m_currentVoxel_c6.m_x3 );
        Point3D backPlane_r2 = new Point3D( m_currentVoxel_c5.m_x1 - m_currentVoxel_c8.m_x1, 
                                            m_currentVoxel_c5.m_x2 - m_currentVoxel_c8.m_x2, 
                                            m_currentVoxel_c5.m_x3 - m_currentVoxel_c8.m_x3 );
        
        m_currentBackVoxelPlane = new Plane3D( m_currentVoxel_c5, backPlane_r1, backPlane_r2 );
        
        // Plane 3: top plane through x2, x3, x6
        
        Point3D topPlane_r1 = new Point3D( m_currentVoxel_c2.m_x1 - m_currentVoxel_c3.m_x1, 
                                           m_currentVoxel_c2.m_x2 - m_currentVoxel_c3.m_x2, 
                                           m_currentVoxel_c2.m_x3 - m_currentVoxel_c3.m_x3 );
        Point3D topPlane_r2 = new Point3D( m_currentVoxel_c2.m_x1 - m_currentVoxel_c6.m_x1, 
                                           m_currentVoxel_c2.m_x2 - m_currentVoxel_c6.m_x2, 
                                           m_currentVoxel_c2.m_x3 - m_currentVoxel_c6.m_x3 );
        
        m_currentTopVoxelPlane = new Plane3D( m_currentVoxel_c2, topPlane_r1, topPlane_r2 );
        
        // Plane 4: bottom plane through x1, x4, x5
        
        Point3D bottomPlane_r1 = new Point3D( m_currentVoxel_c1.m_x1 - m_currentVoxel_c4.m_x1, 
                                              m_currentVoxel_c1.m_x2 - m_currentVoxel_c4.m_x2, 
                                              m_currentVoxel_c1.m_x3 - m_currentVoxel_c4.m_x3 );
        Point3D bottomPlane_r2 = new Point3D( m_currentVoxel_c1.m_x1 - m_currentVoxel_c5.m_x1, 
                                              m_currentVoxel_c1.m_x2 - m_currentVoxel_c5.m_x2, 
                                              m_currentVoxel_c1.m_x3 - m_currentVoxel_c5.m_x3 );
        
        m_currentBottomVoxelPlane = new Plane3D( m_currentVoxel_c1, bottomPlane_r1, bottomPlane_r2 );
    
        // Plane 5: left plane through x1, x2, x5
        
        Point3D leftPlane_r1 = new Point3D( m_currentVoxel_c1.m_x1 - m_currentVoxel_c2.m_x1, 
                                            m_currentVoxel_c1.m_x2 - m_currentVoxel_c2.m_x2, 
                                            m_currentVoxel_c1.m_x3 - m_currentVoxel_c2.m_x3 );
        Point3D leftPlane_r2 = new Point3D( m_currentVoxel_c1.m_x1 - m_currentVoxel_c5.m_x1, 
                                            m_currentVoxel_c1.m_x2 - m_currentVoxel_c5.m_x2, 
                                            m_currentVoxel_c1.m_x3 - m_currentVoxel_c5.m_x3 );
        
        m_currentLeftVoxelPlane = new Plane3D( m_currentVoxel_c1, leftPlane_r1, leftPlane_r2 );
        
        // Plane 6: right plane through x3, x4, x8
        Point3D rightPlane_r1 = new Point3D( m_currentVoxel_c4.m_x1 - m_currentVoxel_c3.m_x1, 
                                             m_currentVoxel_c4.m_x2 - m_currentVoxel_c3.m_x2, 
                                             m_currentVoxel_c4.m_x3 - m_currentVoxel_c3.m_x3 );
        Point3D rightPlane_r2 = new Point3D( m_currentVoxel_c4.m_x1 - m_currentVoxel_c8.m_x1, 
                                             m_currentVoxel_c4.m_x2 - m_currentVoxel_c8.m_x2, 
                                             m_currentVoxel_c4.m_x3 - m_currentVoxel_c8.m_x3 );
        
        m_currentRightVoxelPlane = new Plane3D( m_currentVoxel_c3, rightPlane_r1, rightPlane_r2 );
    }
        
    
    
    /**
     * This method returns whether a spatial 3D point is inside or
     * outside the spectroscopic voxel
     */ 
    public boolean isPointWithinCurrent3DVoxel( Point3D p )
    {
        
        boolean withinVoxel = true;
        
        double frontplane_k     = m_currentFrontVoxelPlane.getNormalPlaneEquationConstant( );
        double backplane_k      = m_currentBackVoxelPlane.getNormalPlaneEquationConstant( );
        double back_point_p_k   = m_currentBackVoxelPlane.computePoint( p );
        
        
        double topplane_k       = m_currentTopVoxelPlane.getNormalPlaneEquationConstant( );
        double bottomplane_k    = m_currentBottomVoxelPlane.getNormalPlaneEquationConstant( );
        double bottom_point_p_k = m_currentBottomVoxelPlane.computePoint( p );
        

        double leftplane_k      = m_currentLeftVoxelPlane.getNormalPlaneEquationConstant( );
        double rightplane_k     = m_currentRightVoxelPlane.getNormalPlaneEquationConstant( );
        double right_point_p_k  = m_currentRightVoxelPlane.computePoint( p );

        if ( !( ( ( frontplane_k <= back_point_p_k ) && ( back_point_p_k <= backplane_k ) ) || 
                ( ( frontplane_k >= back_point_p_k ) && ( back_point_p_k >= backplane_k ) ) ) )
        {
            withinVoxel = false;
        }

        if ( !( ( ( topplane_k <= bottom_point_p_k ) && ( bottom_point_p_k <= bottomplane_k ) ) || 
                ( ( topplane_k >= bottom_point_p_k ) && ( bottom_point_p_k >= bottomplane_k ) ) ) )
        {
            withinVoxel = false;
        }

        if ( !( ( ( leftplane_k <= right_point_p_k ) && ( right_point_p_k <= rightplane_k ) ) || 
                ( ( leftplane_k >= right_point_p_k ) && ( right_point_p_k >= rightplane_k ) ) ) )
        {
            withinVoxel = false;
        }
                
        return withinVoxel;
    }
    
    /**
     * Determines the voxel which is closest to the graphics devices coordinate 
     * (X,Y) and set it selected
     */
    public void setSelectedCSIVoxelFromXYCoordinate( int X, int Y )
    {
        Iterator it = m_gridPointXY.iterator( );
        
        int minDist = Integer.MAX_VALUE;
        
        for ( int k = 0 ; k < this.m_gridSizeDir3; k++ )
        {
            for ( int j = 0 ; j < this.m_gridSizeDir2; j++ )
            {
                for ( int i = 0 ; i < this.m_gridSizeDir1; i++ )
                {
                     Point p = (Point)it.next( );
                     int square = ( p.x - X ) * ( p.x - X ) + ( p.y - Y ) * ( p.y - Y );
                     if ( square  < minDist )
                     {
                         minDist = square;
                         m_selectedVoxelDir1 = i + 1;
                         m_selectedVoxelDir2 = j + 1;
                         m_selectedVoxelDir3 = k;
                     }
                }
            }   
        }
        
        repaint( );
    }
    
    /**
     * Repaints the component
     */
    private void repaint( )
    {
        paintCSIGrid( );
    }
    
    /**
     * Returns the voxel coordinate of the first dimension
     */
    public int getSelectedVoxelDir1( )
    {
        return m_selectedVoxelDir1;
    }
    
    /**
     * Returns the voxel coordinate of the second dimension
     */
    public int getSelectedVoxelDir2( )
    {
        return m_selectedVoxelDir2;
    }

    /**
     * Returns the voxel coordinate of the second dimension
     */
    public int getSelectedVoxelDir3( )
    {
        return m_selectedVoxelDir3;
    }
    
    
    // Attributes CSIGrid
    
    private Point3D m_centerPoint;
    
    private int m_gridSizeDir1, m_gridSizeDir2, m_gridSizeDir3;
    
    private int m_selectedVoxelDir1, m_selectedVoxelDir2, m_selectedVoxelDir3;
    
    private double m_FOVDir1, m_FOVDir2, m_FOVDir3;
    
    private Point3D m_gridDirectionUnitVector1, m_gridDirectionUnitVector2, 
                    m_gridDirectionUnitVector3;
    
    private Point3D m_currentVoxel_c1 = new Point3D( );
    private Point3D m_currentVoxel_c2 = new Point3D( );
    private Point3D m_currentVoxel_c3 = new Point3D( );
    private Point3D m_currentVoxel_c4 = new Point3D( );
    private Point3D m_currentVoxel_c5 = new Point3D( );
    private Point3D m_currentVoxel_c6 = new Point3D( );
    private Point3D m_currentVoxel_c7 = new Point3D( );
    private Point3D m_currentVoxel_c8 = new Point3D( );
    
    
    // 6 planes encapsulating the voxel
    private Plane3D m_currentFrontVoxelPlane;
    private Plane3D m_currentBackVoxelPlane;
    
    private Plane3D m_currentTopVoxelPlane;
    private Plane3D m_currentBottomVoxelPlane;
    
    private Plane3D m_currentLeftVoxelPlane;
    private Plane3D m_currentRightVoxelPlane;
    
    private LinkedList<Point> m_gridPointXY = new LinkedList<Point>( );
    
    private Graphics m_g;
    private Point3D m_p; 
    private Point3D m_r1; 
    private Point3D m_r2; 
    private double m_FOVX; 
    private double m_FOVY;
    private int m_NpixX; 
    private int m_NpixY;
    private String m_comment;
    private String m_title_comment;
}
