/*
 * File: $Id: SimpleThressholdSegmentationCalculator.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 24. September 2004, 11:39 Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2004 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


/*
 * SimpleThressholdSegmentationCalculator.java
 *
 * Created on September 24, 2004, 11:39 AM
 */

package org.spectrim.imageprocessing;


import java.awt.Point;

import java.awt.Cursor;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;


import java.util.Vector;
import java.util.LinkedList;
import java.util.Iterator;

import javax.swing.JFrame;

import org.spectrim.dicom.DicomSliceCanvasPolyDraw;

import org.spectrim.imagegraphicobjects.PolygonUtils;
import org.spectrim.imagegraphicobjects.RegioPolygon;

import org.spectrim.lib.MemTools;
import org.spectrim.lib.Dialogs;
import org.spectrim.lib.SwingWorker;

/**
 *
 * Class that calculates image segments that satifies 
 * pixel value range conditions and maximum vertices
 * number conditions. The major task of this class 
 * is to provide calling classes with ROI lists of 
 * image segments. The ROI's are of type RegioPolygon.
 *
 * @author  hans
 */

public class SimpleThressholdSegmentationCalculator 
{
    /**
     * Creates a new instance of SimpleThressholdSegmentationCalculator
     * @param parent
     * @param image
     */
    public SimpleThressholdSegmentationCalculator( JFrame parent, DicomSliceCanvasPolyDraw image )
    {
        m_parent = parent;
        m_currentView = image;
        
        m_lowThreshold =  0;
        m_highThresshold = Integer.MAX_VALUE;
        
        m_minROISize = 1; 
        m_maxROISize = 100000;
        
        m_segmentationWindow = null;
        
    }
    
    /** 
     * Creates a new instance of SimpleThressholdSegmentationCalculator
     * @param parent
     * @param image
     * @param lowThreshold
     * @param highThresshold
     * @param minROISize
     * @param maxROISize
     */
    public SimpleThressholdSegmentationCalculator( JFrame parent, DicomSliceCanvasPolyDraw image, 
                                                   int  lowThreshold, int highThresshold, 
                                                   int minROISize, int maxROISize ) 
    {
        m_parent = parent;
        m_currentView = image;
        
        m_lowThreshold =  lowThreshold;
        m_highThresshold = highThresshold;
        
        m_minROISize = minROISize; 
        m_maxROISize = maxROISize;
        
        m_segmentationWindow = null;
        
    }
    
    /**
     * Creates a new instance of SimpleThressholdSegmentationCalculator.
     * @param parent
     * @param image
     * @param lowThreshold
     * @param highThresshold
     * @param minROISize
     * @param maxROISize
     * @param segmentationWindow
     */
    public SimpleThressholdSegmentationCalculator( JFrame parent, DicomSliceCanvasPolyDraw image, 
                                                   int  lowThreshold, int highThresshold, 
                                                   int minROISize, int maxROISize,
                                                   RegioPolygon segmentationWindow ) 
    {
        m_parent = parent;
        m_currentView = image;
        
        m_lowThreshold =  lowThreshold;
        m_highThresshold = highThresshold;
        
        m_minROISize = minROISize; 
        m_maxROISize = maxROISize;
        m_segmentationWindow = segmentationWindow;
    }
    
    /**
     * Segments the currently selected view.
     */
    public void segmentCurrentView( ) 
    {
        segmentCurrentView( m_lowThreshold, m_highThresshold );
    }
    
    /**
     * Segments the image in separate thread.
     */
    public void performSegmentation( )
    {
       
       final SwingWorker worker = new SwingWorker( )
       {
            @Override
            public Object construct( )
            {
                try
                {
                    if ( m_segmentationWindow != null )
                    {
                        segmentCurrentView( m_lowThreshold, m_highThresshold, m_segmentationWindow );
                    }
                    else
                    {
                        segmentCurrentView( m_lowThreshold, m_highThresshold );
                    }
                }
                catch ( OutOfMemoryError e )
                {
                    Dialogs d = new Dialogs( );
                    d.messageBox( m_parent, "<HTML> To much ROI's are found to be handled <BR> Change pixel thressholds ranges & roi size ranges </HTML>" );
                }
                return null;
            }
                
            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                m_parent.setCursor( new Cursor( Cursor.DEFAULT_CURSOR ) ); 
                
                notifyActionListeners( PERFORMED_CALCULATION_EVENT );
            }
 
        };
        
        worker.start( );
    }
    
    /**
     * Segments the currently selected view.
     * @param loThreshold
     * @param hiThreshold
     */
    public void segmentCurrentView( int loThreshold, int hiThreshold ) 
    {
	float zoom = m_currentView.getZoom( );
	int w = m_currentView.getImageWidth( );
	int h = m_currentView.getImageHeight( );
	int[][] segVol = MemTools.mallocInt2( h, w );
	
        // Get a reference to the runtime object
        Runtime rt = Runtime.getRuntime( );
        
        m_currentView.deleteAllPolygonsOfCurrentlyDisplayedImageOnly( );
        // force garbage collection
        rt.gc( );
        
        if ( segVol == null ) 
        {
	    Dialogs.messageBox( m_parent, "Not enough memory for operation" );
	    return;
	}

	// label the pixels

	//System.out.println( "Labeling pixels ..." );

	for ( int k = 1; k < h - 1; k++ ) 
        {
	    int off = k * w;
	    for ( int l = 1; l < w - 1; l++ )
            {
		int pix = m_currentView.getRawData( )[off++];
		if ( pix >= loThreshold && pix <= hiThreshold )
		    segVol[k][l] = -1;
	    }
	}
	
	//System.out.println( "creating borders ..." );
        
        int added = 0;
        int rejected = 0;
        
	int polygonIndex = 1;
	Point startingPoint = new Point( 0, 0 );
        PolygonUtils pUtils = new PolygonUtils( );
	while ( null != ( startingPoint = pUtils.getPolygonStart( segVol, startingPoint ) ) ) 
        {
	    RegioPolygon P = pUtils.regionBorder( segVol, startingPoint, polygonIndex );
	    P.setScalingFactor( zoom );
	    // P.reduceVertices( 40.0 );
            
            if ( ( P.getInnerVolume( ) > m_minROISize ) && ( P.getInnerVolume( ) < m_maxROISize ) )
            {
                m_currentView.addPolygon( P );
                added++;
            }
            else
            {
                P = null;
                rejected++;
                rt.gc( );
            }
	    polygonIndex++;
	}

        applyRegioPolygonSizeRequirements( );
        
        m_currentView.repaint( );
        
        segVol = null;
        rt.gc( );
        
        System.out.println( "Added : " + added + " Rejected : " + rejected );
        
    }
    

    /**
     * Computes the RegioPolygons of given imageNumber and overlayNumber.
     * @param imageNumber
     * @param overlayNumber
     * @param deleteExistingPolygons
     */
    public void computeRegioPolygonObjectsOfSegmentOverlayOfView( int imageNumber, int overlayNumber, boolean deleteExistingPolygons ) 
    {
        float zoom = m_currentView.getZoom( );
        int w = m_currentView.getImageWidth( );
        int h = m_currentView.getImageHeight( );
        int[][] segVol = MemTools.mallocInt2( h, w );
	
        // Get a reference to the runtime object
        Runtime rt = Runtime.getRuntime( );
        if ( deleteExistingPolygons == true )
        {
            m_currentView.deleteAllPolygonsOfCurrentlyDisplayedImageOnly( );
        }
        // force garbage collection
        rt.gc( );
        
        if ( segVol == null ) 
        {
            Dialogs.messageBox( m_parent, "Not enough memory for operation" );
            return;
        }

        // get the image segment overlay from the DicomSliceCanvasPolydraw
        boolean[][] overlay =  m_currentView.getImageSegmentOverlayLinkedLists( ).get( imageNumber ).get( overlayNumber );

        // label the pixels

        //System.out.println( "Labeling pixels ..." );

        for ( int k = 1; k < h - 1; k++ )
        {
            int off = k * w;
            for ( int l = 1; l < w - 1; l++ )
            {
                boolean pix = overlay[l][k];
                if ( pix == false )
                {
                    segVol[k][l] = -1;
                }
                else
                {
                    segVol[k][l] = 0;
                }
            }
        }
	
        //System.out.println( "creating borders ..." );
        
        int added = 0;
        int rejected = 0;
        
        int polygonIndex = 1;
        Point startingPoint = new Point( 0, 0 );
        PolygonUtils pUtils = new PolygonUtils( );
        while ( null != ( startingPoint = pUtils.getPolygonStart( segVol, startingPoint ) ) )
        {
            RegioPolygon P = pUtils.regionBorder( segVol, startingPoint, polygonIndex );
            P.setScalingFactor( zoom );
            // P.reduceVertices( 40.0 );
            P.setLabel( "roi_" + imageNumber + "_"+ overlayNumber + "_" + added );
            if ( ( P.getInnerVolume( ) > m_minROISize ) && ( P.getInnerVolume( ) < m_maxROISize ) )
            {
                m_currentView.addPolygon( imageNumber, P );
                added++;
            }
            else
            {
                rejected++;
                P = null;
                rt.gc( );
            }
            polygonIndex++;
        }

        applyRegioPolygonSizeRequirements( );
        
        m_currentView.repaint( );
        
        segVol = null;
        rt.gc( );
        
        System.out.println( "Added : " + added + " Rejected : " + rejected );
        
    }
    
    
    /**
     * Segments the currently selected view within the segment window only.
     * @param loThreshold
     * @param hiThreshold
     * @param segmentWindow
     */
    public void segmentCurrentView( int loThreshold, int hiThreshold, RegioPolygon segmentWindow ) 
    {
	float zoom = m_currentView.getZoom( );
	int w = m_currentView.getImageWidth( );
	int h = m_currentView.getImageHeight( );
	int[][] segVol = MemTools.mallocInt2( h, w );
	RegioPolygon segMentWin = new RegioPolygon( segmentWindow );
        
        segMentWin.setScalingFactor( 1f / zoom );
        
        // Get a reference to the runtime object
        Runtime rt = Runtime.getRuntime( );
        
        m_currentView.deleteAllPolygonsOfCurrentlyDisplayedImageOnly( );
        // force garbage collection
        rt.gc( );
        
        if ( segVol == null ) 
        {
	    Dialogs.messageBox( m_parent, "Not enough memory for operation" );
	    return;
	}

	// label the pixels

	//System.out.println( "Labeling pixels ..." );

	for ( int k = 1; k < h - 1; k++ ) 
        {
	    int off = k * w;
	    for ( int l = 1; l < w - 1; l++ )
            {
		int pix = m_currentView.getRawData( )[off++];
		if ( ( pix >= loThreshold && pix <= hiThreshold ) && ( segMentWin.contains( l, k )== true ) )
		    segVol[k][l] = -1;
	    }
	}
        
        
	// System.out.println( "creating borders ..." );
        
        int added = 0;
        int rejected = 0;
        
	int polygonIndex = 1;
	Point startingPoint = new Point( );
        PolygonUtils pUtils = new PolygonUtils( );
	while ( null != ( startingPoint = pUtils.getPolygonStart( segVol, startingPoint ) ) ) 
        {
	    RegioPolygon P = pUtils.regionBorder( segVol, startingPoint, polygonIndex );
	    P.setScalingFactor( zoom );
	    // P.reduceVertices( 40.0 );
            
            if ( ( P.getInnerVolume( ) > m_minROISize ) && ( P.getInnerVolume( ) < m_maxROISize ) )
            {
                m_currentView.addPolygon( P );
                added++;
            }
            else
            {
                P = null;
                rejected++;
                rt.gc( );
            }
	    polygonIndex++;
	}

        applyRegioPolygonSizeRequirements( );
        
        m_currentView.repaint( );
        
        segVol = null;
        rt.gc( );
        
        System.out.println( "Added : " + added + " Rejected : " + rejected );
        
    }
    
    
    /**
     * Computes the RegioPolygons the currently selected view.
     * @param imageNumber
     * @param overlayNumber
     * @param loThreshold
     * @param hiThreshold
     * @param segmentWindow
     */
    public void computeRegioPolygonObjectsOfSegmentOverlayOfView( int imageNumber, int overlayNumber,
                                                                  int loThreshold, int hiThreshold, 
                                                                  RegioPolygon segmentWindow ) 
    {
        
        // look above for correctly coding 4-12-2008
	float zoom = m_currentView.getZoom( );
	int w = m_currentView.getImageWidth( );
	int h = m_currentView.getImageHeight( );
	int[][] segVol = MemTools.mallocInt2( h, w );
	RegioPolygon segMentWin = new RegioPolygon( segmentWindow );
        
        segMentWin.setScalingFactor( 1f / zoom );
        
        // Get a reference to the runtime object
        Runtime rt = Runtime.getRuntime( );
        
        m_currentView.deleteAllPolygonsOfCurrentlyDisplayedImageOnly( );
        // force garbage collection
        rt.gc( );
        
        if ( segVol == null ) 
        {
	    Dialogs.messageBox( m_parent, "Not enough memory for operation" );
	    return;
	}
        
        // get the image segment overlay from the DicomSliceCanvasPolydraw
        boolean[][] overlay =  m_currentView.getImageSegmentOverlayLinkedLists( ).get( imageNumber ).get(overlayNumber);
        
        
	// label the pixels

	//System.out.println( "Labeling pixels ..." );

	for ( int k = 1; k < h - 1; k++ ) 
        {
	    int off = k * w;
	    for ( int l = 1; l < w - 1; l++ )
            {
		boolean pix = overlay[l][k];
		if ( pix == false )
                {
		    segVol[k][l] = -1;
                }
	    }
	}
        
	// System.out.println( "creating borders ..." );
        
        int added = 0;
        int rejected = 0;
        
	int polygonIndex = 1;
	Point startingPoint = new Point( );
        PolygonUtils pUtils = new PolygonUtils( );
	while ( null != ( startingPoint = pUtils.getPolygonStart( segVol, startingPoint ) ) ) 
        {
	    RegioPolygon P = pUtils.regionBorder( segVol, startingPoint, polygonIndex );
	    P.setScalingFactor( zoom );
	    // P.reduceVertices( 40.0 );
            
            if ( ( P.getInnerVolume( ) > m_minROISize ) && ( P.getInnerVolume( ) < m_maxROISize ) )
            {
                m_currentView.addPolygon( P );
                added++;
            }
            else
            {
                P = null;
                rejected++;
                rt.gc( );
            }
	    polygonIndex++;
	}

        applyRegioPolygonSizeRequirements( );
        
        m_currentView.repaint( );
        
        segVol = null;
        rt.gc( );
        
        System.out.println( "Added : " + added + " Rejected : " + rejected );
        
    }
    
    
    /**
     * Segmentation only allows for RegioPolygons to have a specific minimum and maximum
     * size all regio's not in this range are removed from the list.
     */
    private void applyRegioPolygonSizeRequirements( )
    {
        Vector<RegioPolygon> removeList = new Vector<RegioPolygon>( );
	
        int n = m_currentView.getNumberOfPolygons( );
	
        for ( int k = 0; k < n; k++ ) 
        {
            RegioPolygon p = m_currentView.getPolygon( k );
            if ( ( !p.isBiggerThan( m_minROISize ) ) || ( p.isBiggerThan( m_maxROISize ) ) ) 
            {
                removeList.addElement( p );
            }
        }
        
        n = removeList.size( );
        
        for ( int k = 0; k < n; k++ )
        {
            m_currentView.deletePoly( removeList.elementAt( k ) );
        }
    }
    
    
    /**
     * Adds an action listener to the list of action listeners 
     * @param e
     */
    public void addActionListener( ActionListener e )
    {
	m_listeners.add( e );
    }
    
    /**
     * Notify all actionListeners.
     */
    private void notifyActionListeners( int eventType )
    {
        Iterator it = m_listeners.iterator( );
        while( it.hasNext( ) )
        {
            ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( this, eventType, "" ) );
        }
    }
    
    
    // Private attributes of this class
    private JFrame m_parent;
    private DicomSliceCanvasPolyDraw m_currentView;
    
    private static int m_lowThreshold =  0;
    private static int m_highThresshold = 0;
    
    private int m_minROISize = 0;
    private int m_maxROISize = 0;
    
    private RegioPolygon m_segmentationWindow;
    
    // Listeners to changes in the segmentation gui control
    private LinkedList<ActionListener> m_listeners = new LinkedList<>( );
    
    public final static int PERFORMED_CALCULATION_EVENT = 0;
    
}
