/*
 * File: $Id: MRParameterSpectrumModelViewer.java,v 1.1.1.1 2006/04/19 09:38:49 hans Exp $
 * History
 * Date: 08.07.2003, 09:14  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


package org.spectrim.tdfdfit;

import org.spectrim.xyplot.XYPlotRange;

import java.awt.Cursor;
import java.awt.Color;
import java.awt.Point;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;

import java.awt.geom.Point2D;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import java.util.LinkedList; 
import java.util.Iterator;

import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;

import java.awt.print.Book;
import java.awt.print.Printable;
import java.awt.print.PageFormat;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import javax.swing.JOptionPane;
import org.spectrim.tdfdfit.quantitativeresults.QuantitativeSVSContextDataContainer;
import org.spectrim.tdfdfit.waterreferencemodelling.TwoCompartmentWaterReferenceModelFittingFrame;
import org.spectrim.dicom.DicomDirectory;


/**
 * Specialized class derived from SingleVoxelSpectrumViewer that allows for
 * display and manipulation of spectral models.
 *
 * @author  hans
 */
public class MRParameterSpectrumModelViewer extends SingleVoxelSpectrumViewer implements ActionListener, MouseListener, MouseMotionListener, Printable
{
    
    /**
     * Constructor needed to be able to use the component as Java Bean.
     */
    public MRParameterSpectrumModelViewer( )
    {
        super( );
        setDisplayXReversed( true );
    }
    
    /** 
     * Creates a new instance of ParameterSpectrumModelViewer. 
     * @param title
     * @param parent
     * @param ddir
     */
    public MRParameterSpectrumModelViewer( String title, JFrame parent, 
                                           DicomDirectory ddir ) 
    {
        super( title, parent, ddir );
        
        if ( this.getMouseListeners( ) == null )
        {
            addMouseListener( this );
            addMouseMotionListener( this );
        }
        
        // This class adds functionality to the popupmenu of the base classes
        addPopupMenuItems( );
        
        TDFDFittingDefaults def = TDFDFittingDefaults.getInstance( );
        def.load( );
        
        if ( def.getProperty( TDFDFittingDefaults.DISPLAY_UNIT ).compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 )
        {
            this.setDisplayXReversed( true );
        }
        else
        {
            this.setDisplayXReversed( false );
        }
    }
    
    /**
     * Adds entry point for user interactive spectrum model editing.
     */
    private void addPopupMenuItems( )
    {
        m_modelSpectumEditing = new JMenu( "Spectrum model editing" );
        
        m_addLineMenuItem = new JMenuItem( "Add line" );
        m_addLineMenuItem.addActionListener( this );
        m_modelSpectumEditing.add( m_addLineMenuItem );
        
        m_popupmenu.add( m_modelSpectumEditing );
        m_popupmenu.addSeparator( );
        
        m_applyLineShapeTransform = new JMenuItem( "Apply lineshape transform (QUALITY) .." );
        m_applyLineShapeTransform.setToolTipText( "Applies QUALITY to the measured spectrum" );
        m_applyLineShapeTransform.addActionListener( this );
        
        m_popupmenu.add( m_applyLineShapeTransform );
        
        m_printSpectrum = new JMenuItem( "Print" );
        m_printSpectrum.addActionListener( this );
        
        m_popupmenu.add( m_printSpectrum );
        
    }
    
    /**
     * Draws the spectra, with the response of the selected node in red.
     */
    private void drawSpectraAccordingSelectedNode( )
    {
        TDFDFitPrototypeModelSpectrumParameters modSpec = TDFDFitPrototypeModelSpectrumParameters.getInstance( );
        m_responseList.clear( );
        m_responseNameList.clear( );
        m_responseListBuffer.clear( );
        deleteAllCurves( );
        if ( m_measuredSpectrum !=  null )
        {
            addSpectralDataSet( "measured", m_measuredSpectrum, false );
        }
        
        MRLineModelSpectralResponse completeModel = modSpec.getCompleteModelResponse( );
        if ( completeModel != null )
        {
            addSpectralDataSet( "model", completeModel,  false );
        }
        
        MRLineModelSpectralResponse partialResponse = modSpec.getSelectedModelResponse( );
        if ( partialResponse != null )
        {
            addSpectralDataSet( "part", partialResponse,  false );
        }
        
        // Force a draw of the newly set data 
        m_firstTimeChanged = true;
        changeDisplayedSignal( );
    }
    
    /**
     * Called whenever an ActionEvent occurs.
     */
    @Override
    public void actionPerformed( ActionEvent e )
    {
        
        Object sender = e.getSource( );
        
        if ( sender instanceof String )
        {
            // update only if a valid model is defined
            if ( ( (String)sender).compareTo( "No model defined" ) != 0 )
            {    
                drawSpectraAccordingSelectedNode( );
            }
        }
        
        if ( sender instanceof TDFDFitPrototypeModelSpectrumParameters )
        {
            // System.out.println( "Print e-getID( ) = " + e.getID( ) );
            if ( e.getID( ) == TDFDFitPrototypeModelSpectrumParameters.MODEL_CHANGED )
            {
                drawSpectraAccordingSelectedNode( );
            }
        }
        
        if ( sender == m_addLineMenuItem )
        {
            m_selectXYPointTriggerSet = true;
            this.setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
        }
        
        
        // if JRadioButtonMenuItem is selected than  
        if ( sender == m_applyLineShapeTransform )
        {
            QuantitativeSVSContextDataContainer waterRefData = QuantitativeSVSContextDataContainer.getInstance( );
            if ( waterRefData.getSimpleWaterRefefenceSignalSet( ) == false )
            {
                JOptionPane.showMessageDialog( null, 
                                               "You are now prompted to define this/these signals", 
                                               "No water references signal(s) defined.", 
                                               JOptionPane.ERROR_MESSAGE );
                
                m_waterReferenceFrame = new TwoCompartmentWaterReferenceModelFittingFrame( null, m_dicomDir );
            }
            else
            {
                double[] waterRefReal = waterRefData.getWaterReferenceSignalReal( );
                double[] waterRefImag = waterRefData.getWaterReferenceSignalImag( );
                
                getMeasuredSpectrum( ).applyQuality( true, waterRefReal, waterRefImag );
                changeDisplayedSignal( );
            }
        }
        
        if ( sender == this.m_printSpectrum )
        {
            print( );
        }
        
        super.actionPerformed( e );
    }
    
    public void expectMousePressedModeForInteractiveVoigtLineDefinition( )
    {
        m_selectXYPointTriggerSet = true;
        this.setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
    }
    
    /**
     * 
     * Adds an ActionListener to the list of action listeners who are notified
     * when something happens with the spin box.
     * 
     * @param e
     */
    @Override
    public void addActionListener( ActionListener e )
    {
	m_listeners.add( e );
        super.addActionListener( e );
    }

    /**
     * 
     * Called when the mouse button is pressed.
     * @param mouseEvent
     * 
     */
    @Override
    public void mousePressed( MouseEvent mouseEvent ) 
    {
       
        // is only of interest if left button is pressed
        if ( ( mouseEvent.getModifiers( ) & InputEvent.BUTTON1_MASK ) == InputEvent.BUTTON1_MASK )
        {
            m_mouseButtonPressedX = mouseEvent.getX( );
            m_mouseButtonPressedY = mouseEvent.getY( );
        
            if ( this.m_selectXYPointTriggerSet == true )
            {      
                // determineValuesAtCursorPosition( m_mouseButtonPressedX, m_mouseButtonPressedY );
                Point2D.Float pnt2D = getXYCoordinatesAtCurrentMousePositionInCorrectUnits( m_mouseButtonPressedX, m_mouseButtonPressedY );
                this.setCursor( Cursor.getPredefinedCursor(  Cursor.DEFAULT_CURSOR ) );
                notifyActionListeners( ADD_LINE_EVENT, "add user indicated line" );
                m_selectXYPointTriggerSet = false;
            }
        }
        
        super.mousePressed( mouseEvent );
    }
    
    /**
     * Called when the mouse button is pressed.
     * @param mouseEvent
     */
    @Override
    public void mouseReleased( MouseEvent mouseEvent )
    {

        if ( m_mouseDragged == true )
        {
            m_mouseDragged = false;
            // is only of interest if left button is released
            if ( ( mouseEvent.getModifiers( ) & InputEvent.BUTTON1_MASK ) == InputEvent.BUTTON1_MASK )
            {
            
                m_mouseButtonReleasedX = mouseEvent.getX( );
                m_mouseButtonReleasedY = mouseEvent.getY( );
            
                // continue only if the start and stop position are not the same
                if ( ( Math.abs( m_mouseButtonPressedX - m_mouseButtonReleasedX ) > 5 ) && 
                     ( Math.abs( m_mouseButtonPressedY - m_mouseButtonReleasedY ) > 5 ) )
                {
                    Point2D.Double start = determineValuesAtCursorPosition( new Point( m_mouseButtonPressedX, m_mouseButtonPressedY ) );
                    Point2D.Double stop = determineValuesAtCursorPosition( new Point( m_mouseButtonReleasedX, m_mouseButtonReleasedY ) );
                    
                    float startX = getXPositionOfCurrentRepresentation( m_mouseButtonPressedX, m_mouseButtonPressedY );
                    float stopX = getXPositionOfCurrentRepresentation( m_mouseButtonReleasedX, m_mouseButtonReleasedY );
                    
                    
                    if ( ( startX < stopX ) && ( start.y < stop.y ) )
                    {
                        m_xyForcedPlotRange = new XYPlotRange( (float)startX, (float)stopX, (float)start.y, (float)stop.y );
                    }
                    
                    if ( ( startX < stopX ) && ( start.y > stop.y ) )
                    {
                        m_xyForcedPlotRange = new XYPlotRange( (float)startX, (float)stopX, (float)stop.y, (float)start.y );   
                    }
                    
                    if ( ( startX > stopX ) && ( start.y > stop.y ) )
                    {
                        m_xyForcedPlotRange = new XYPlotRange( (float)stopX, (float)startX, (float)stop.y, (float)start.y );
                    }
                    
                    if ( ( startX > stopX ) && ( start.y < stop.y ) )
                    {
                        m_xyForcedPlotRange = new XYPlotRange( (float)stopX, (float)startX, (float)start.y, (float)stop.y );
                    }
                    
                    m_XYPlotRangeSelectorMenuItem.setSelected( true );
                    
                    changeDisplayedSignal( );
                    notifyActionListeners( USER_REQUESTED_XYPLOT_ZOOM_EVENT , USER_REQUESTED_XYPLOT_ZOOM_EVENT_CMD );
                }
            }
        }
        
        super.mouseReleased( mouseEvent );
    }
    
    /**
     * Called when the mouse button is clicked.
     * @param mouseEvent
     */
    @Override
    public void mouseClicked( MouseEvent mouseEvent ) 
    {
        // is only of interest if left button is pressed
        if ( ( mouseEvent.getModifiers( ) & InputEvent.BUTTON1_MASK ) == InputEvent.BUTTON1_MASK )
        {
            m_mouseButtonClickedX = mouseEvent.getX( );
            m_mouseButtonClickedY = mouseEvent.getY( );
        
            if ( this.m_selectXYPointTriggerSet == true )
            {     
                    
                Point2D.Float pnt = getXYCoordinatesAtCurrentMousePositionInCorrectUnits( m_mouseButtonClickedX, m_mouseButtonClickedY );
                
                this.m_lastSelectedAmplitudeValueAtMouseClickEvent = pnt.y;
                this.m_lastSelectedOffsetValueAtMouseClickEvent = pnt.x;
                
                // determineValuesAtCursorPosition( m_mouseButtonClickedX, m_mouseButtonClickedY );
            
                this.setCursor( Cursor.getPredefinedCursor( Cursor.DEFAULT_CURSOR ) );
                notifyActionListeners( ADD_LINE_EVENT, "add user indicated line" );
                m_selectXYPointTriggerSet = false;
            }
        }
        super.mouseClicked( mouseEvent );
    }

    /**
     * Called when the mouse is dragged.
     * @param mouseEvent
     */
    @Override
    public void mouseDragged( MouseEvent mouseEvent )
    {
        m_mouseDragged = true;
        super.mouseDragged( mouseEvent );
    }
    
    /**
     * Called when the mouse enters the control.
     * @param mouseEvent
     */    
    @Override
    public void mouseEntered( MouseEvent mouseEvent )
    {
        super.mouseEntered( mouseEvent );
    }
    
    /**
     * Called when the mouse exits.
     * @param mouseEvent
     */    
    @Override
    public void mouseExited( MouseEvent mouseEvent )
    {
        super.mouseExited( mouseEvent );
    }
    
    /**
     * Called when the mouse button is released.
     * @param mouseEvent
     */    
    @Override
    public void mouseMoved( MouseEvent mouseEvent )
    {
        createStatusLineMessageMouseLocation( mouseEvent );
        super.mouseMoved( mouseEvent );
    }
    
    /**
     * Creates a status line containing the current location. 
     * of the mouse pointer.
     */
    private void createStatusLineMessageMouseLocation( MouseEvent mouseEvent )
    {
        
        if ( m_statusPanel != null )
        {
            int xPos = mouseEvent.getX( );
            int yPos = mouseEvent.getY( );

            if ( ( xPos > this.getSize( ).getWidth( ) / 10 ) && ( xPos < 9 * this.getSize( ).getWidth( ) / 10 ) )
            {   
                if ( ( yPos > this.getSize( ).getHeight( ) / 10 ) && ( yPos < 9 * this.getSize( ).getHeight( ) / 10 ) )
                {
                    float xrange = 8.0f * (float)getSize( ).getWidth( ) / 10.0f;
                    float yrange = 8.0f * (float)getSize( ).getHeight( ) / 10.0f;

                    MRSignalProperties mrcond = MRSignalProperties.getInstance( );

                    if ( m_curves != null )
                    {
                        if ( m_curves.size( ) > 0 )
                        {
                            float xPosCorrected = ( ( (float)xPos - (float)this.getSize( ).getWidth( ) / 10.0f )  )  / xrange;
                            float yPosCorrected = ( ( (float)yPos - (float)this.getSize( ).getHeight( ) / 10.0f )  )  / yrange;

                            if ( m_XYPlotRangeSelectorMenuItem.isSelected( ) == true )
                            {
                                if ( this.m_displayXReversed == false )
                                {
                                    xPosCorrected = m_xyForcedPlotRange.getMinimumX( ) + xPosCorrected * m_xyForcedPlotRange.getXRange( );
                                }
                                else
                                {
                                    xPosCorrected = m_xyForcedPlotRange.getMaximumX( ) - xPosCorrected * m_xyForcedPlotRange.getXRange( );
                                }

                                yPosCorrected = m_xyForcedPlotRange.getMaximumY( ) - yPosCorrected * m_xyForcedPlotRange.getYRange( );
                            }
                            else
                            {
                                if ( this.m_displayXReversed == false )
                                {
                                    xPosCorrected = m_xyAutoPlotRange.getMinimumX( ) + xPosCorrected * m_xyAutoPlotRange.getXRange( );
                                }
                                else
                                {
                                    xPosCorrected = m_xyAutoPlotRange.getMaximumX( ) - xPosCorrected * m_xyAutoPlotRange.getXRange( );
                                }

                                yPosCorrected = m_xyAutoPlotRange.getMaximumY( ) - yPosCorrected * m_xyAutoPlotRange.getYRange( );
                            }

                            m_statusPanel.setLeftMessage( "( x, y ) =  ( " + xPosCorrected + " , " + yPosCorrected + " )" );
                        }
                    }
                }
            }
        }
    }
    
    
    /**
     * Creates a status line containing the current location. 
     * of the mouse pointer.
     */
    protected Point2D.Float getXYCoordinatesAtCurrentMousePositionInCorrectUnits( int xPixelPos, int yPixelPos )
    {
        
        if ( m_statusPanel != null )
        {
//            int xPos = mouseEvent.getX( );
//            int yPos = mouseEvent.getY( );

            if ( ( xPixelPos >= (int)this.getSize( ).getWidth( ) / 10 ) && ( xPixelPos <= (int)(9 * this.getSize( ).getWidth( ) / 10 ) ) )
            {   
                if ( ( yPixelPos >= (int)this.getSize( ).getHeight( ) / 10 ) && ( yPixelPos <= (int)(9 * this.getSize( ).getHeight( ) / 10 ) ) )
                {
                    float xrange = 8.0f * (float)getSize( ).getWidth( ) / 10.0f;
                    float yrange = 8.0f * (float)getSize( ).getHeight( ) / 10.0f;

                    if ( m_curves != null )
                    {
                        if ( m_curves.size( ) > 0 )
                        {
                            float xPosCorrected = ( ( (float)xPixelPos - (float)this.getSize( ).getWidth( ) / 10.0f )  )  / xrange;
                            float yPosCorrected = ( ( (float)yPixelPos - (float)this.getSize( ).getHeight( ) / 10.0f )  )  / yrange;

                            if ( m_XYPlotRangeSelectorMenuItem.isSelected( ) == true )
                            {
                                if ( this.m_displayXReversed == false )
                                {
                                    xPosCorrected = m_xyForcedPlotRange.getMinimumX( ) + xPosCorrected * m_xyForcedPlotRange.getXRange( );
                                }
                                else
                                {
                                    xPosCorrected = m_xyForcedPlotRange.getMaximumX( ) - xPosCorrected * m_xyForcedPlotRange.getXRange( );
                                }

                                yPosCorrected = m_xyForcedPlotRange.getMaximumY( ) - yPosCorrected * m_xyForcedPlotRange.getYRange( );
                            }
                            else
                            {
                                if ( this.m_displayXReversed == false )
                                {
                                    xPosCorrected = m_xyAutoPlotRange.getMinimumX( ) + xPosCorrected * m_xyAutoPlotRange.getXRange( );
                                }
                                else
                                {
                                    xPosCorrected = m_xyAutoPlotRange.getMaximumX( ) - xPosCorrected * m_xyAutoPlotRange.getXRange( );
                                }

                                yPosCorrected = m_xyAutoPlotRange.getMaximumY( ) - yPosCorrected * m_xyAutoPlotRange.getYRange( );
                            }

                            // System.out.println( "Real units at curson click position ( x, y ) = ( " + xPosCorrected + " , " + yPosCorrected + " )" );
                           
                            
                            m_lastSelectedOffsetValueAtMouseClickEvent = xPosCorrected;
                            m_lastSelectedAmplitudeValueAtMouseClickEvent = yPosCorrected;
                        }
                    }
                }
            }
        }
        
        return new Point2D.Float( m_lastSelectedOffsetValueAtMouseClickEvent, 
                                  m_lastSelectedAmplitudeValueAtMouseClickEvent );
        
    }
    
    /**
     * Method to compute the forced plot range (if user zoomed the plot panel
     * and the horizontal frequency display units are changed.
     */
    public void updateForcedXYPlotRange( String previouslyDisplayType, String currentDisplayType )
    {
        
        if ( m_XYPlotRangeSelectorMenuItem.isSelected( ) == true )
        {

            // Current forced plot range
            System.out.println( "Previous display type : " + previouslyDisplayType );
            System.out.println( "Current display type  : " + currentDisplayType );

            System.out.println( "Minimum X : " + m_xyForcedPlotRange.getMinimumX( ) );
            System.out.println( "Maximum X : " + m_xyForcedPlotRange.getMaximumX( ) );
            System.out.println( "X-range   : " + m_xyForcedPlotRange.getXRange( ) );

//            // System.out.println( "X-range   : " + TDFDFittingDefaults.getInstance( ).getProperty( ) );
//            Point2D.Float oldMinPoint = getXYCoordinatesAtCurrentMousePositionInCorrectUnits( (int)(this.getSize( ).getWidth( ) / 10 ), 
//                                                                                              (int)(this.getSize( ).getHeight( ) / 10 ) );
//            Point2D.Float oldMaxPoint = getXYCoordinatesAtCurrentMousePositionInCorrectUnits( (int)(9 * this.getSize( ).getWidth( ) / 10 ), 
//                                                                                              (int)(this.getSize( ).getHeight( ) / 10 ) );
            double minX   = m_xyForcedPlotRange.getMinimumX( );
            double maxX   = m_xyForcedPlotRange.getMaximumX( );
            double rangeX = maxX - minX;

            
            // points => Hz
            if ( ( previouslyDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 ) &&  
                 ( currentDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 ) )

            {
                minX   =  m_xyForcedPlotRange.getMinimumX( ) * MRSignalProperties.getInstance( ).getAcquistitionBandwidth( ) / MRSignalProperties.getInstance( ).getNumberOfSamplePoints( );
                maxX   =  m_xyForcedPlotRange.getMaximumX( ) * MRSignalProperties.getInstance( ).getAcquistitionBandwidth( ) / MRSignalProperties.getInstance( ).getNumberOfSamplePoints( );
                
            }
            
            // Hz => points
            if ( ( previouslyDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 ) &&  
                 ( currentDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 ) )

            {
                minX   =  m_xyForcedPlotRange.getMinimumX( ) * MRSignalProperties.getInstance( ).getNumberOfSamplePoints( ) / MRSignalProperties.getInstance( ).getAcquistitionBandwidth( );
                maxX   =  m_xyForcedPlotRange.getMaximumX( ) * MRSignalProperties.getInstance( ).getNumberOfSamplePoints( ) / MRSignalProperties.getInstance( ).getAcquistitionBandwidth( );
                
            }

            // points => ppm
            if ( ( previouslyDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 ) &&  
                 ( currentDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 ) )

            {
                
                MRSignalProperties props = MRSignalProperties.getInstance( );
                float ppmRange = props.getAcquistitionBandwidth( ) / ( props.getResonanceFrequency( ) / 1000000f );
                
                minX = - ( ppmRange / props.getNumberOfSamplePoints( ) ) * minX +  props.getPPMValueInCenterOfSpectrum( );
                maxX = - ( ppmRange / props.getNumberOfSamplePoints( ) ) * maxX +  props.getPPMValueInCenterOfSpectrum( );
                
            }

            // ppm => points
            if ( ( previouslyDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 ) &&  
                 ( currentDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 ) )

            {
                MRSignalProperties props = MRSignalProperties.getInstance( );
                float ppmRange = props.getAcquistitionBandwidth( ) / ( props.getResonanceFrequency( ) / 1000000f );
                 
                minX = - ( props.getNumberOfSamplePoints( ) / ppmRange ) * minX + (  props.getNumberOfSamplePoints( ) / ppmRange ) * props.getPPMValueInCenterOfSpectrum( );
                maxX = - ( props.getNumberOfSamplePoints( ) / ppmRange ) * maxX + (  props.getNumberOfSamplePoints( ) / ppmRange ) * props.getPPMValueInCenterOfSpectrum( );
            }

            
            // Hz => ppm
            if ( ( previouslyDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 ) &&  
                 ( currentDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 ) )

            {
                
                MRSignalProperties props = MRSignalProperties.getInstance( );
                float ppmRange = props.getAcquistitionBandwidth( ) / ( props.getResonanceFrequency( ) / 1000000f );
                
                minX = -( ppmRange/props.getAcquistitionBandwidth( ) ) * minX + props.getPPMValueInCenterOfSpectrum( );
                maxX = -( ppmRange/props.getAcquistitionBandwidth( ) ) * maxX + props.getPPMValueInCenterOfSpectrum( );
            }

            // ppm => Hz
            if ( ( previouslyDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 ) &&  
                 ( currentDisplayType.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 ) )

            {
                MRSignalProperties props = MRSignalProperties.getInstance( );
                float ppmRange = props.getAcquistitionBandwidth( ) / ( props.getResonanceFrequency( ) / 1000000f );
                 
                minX = - ( props.getAcquistitionBandwidth( ) / ppmRange ) * minX + (  props.getAcquistitionBandwidth( ) / ppmRange ) * props.getPPMValueInCenterOfSpectrum( );
                maxX = - ( props.getAcquistitionBandwidth( ) / ppmRange ) * maxX + (  props.getAcquistitionBandwidth( ) / ppmRange ) * props.getPPMValueInCenterOfSpectrum( );
            }
            
            if ( minX > maxX )
            {
                m_xyForcedPlotRange.setMinimumX( (float)maxX );
                m_xyForcedPlotRange.setMaximumX( (float)minX );

            }
            else
            { 
                m_xyForcedPlotRange.setMinimumX( (float)minX );
                m_xyForcedPlotRange.setMaximumX( (float)maxX );
            }
            
            
            
            
        }
    }
    
    /**
     * 
     * Returns the x-axis value in the current representation.
     * 
     * @param currentX
     * @param currentY
     * @return 
     */
    public float getXPositionOfCurrentRepresentation( int currentX, int currentY )
    {
        float xPosCorrected = -1f;
        float yPosCorrected = -1f;
        
        if ( ( currentX >= this.getSize( ).getWidth( ) / 10 ) && ( currentX <= 9 * this.getSize( ).getWidth( ) / 10 ) )
        {   
            if ( ( currentY >= this.getSize( ).getHeight( ) / 10 ) && ( currentY <= 9 * this.getSize( ).getHeight( ) / 10 ) )
            {
                float xrange = 8.0f * (float)getSize( ).getWidth( ) / 10.0f;
                float yrange = 8.0f * (float)getSize( ).getHeight( ) / 10.0f;
                
                MRSignalProperties mrcond = MRSignalProperties.getInstance( );
                
                if ( m_curves != null )
                {
                    if ( m_curves.size( ) > 0 )
                    {
                        xPosCorrected = ( ( (float)currentX - (float)this.getSize( ).getWidth( ) / 10.0f )  )  / xrange;
                        yPosCorrected = ( ( (float)currentY - (float)this.getSize( ).getHeight( ) / 10.0f )  )  / yrange;
                        
                        if ( m_XYPlotRangeSelectorMenuItem.isSelected( ) == true )
                        {
                            if ( m_displayXReversed == false )
                            {
                                xPosCorrected = m_xyForcedPlotRange.getMinimumX( ) + xPosCorrected * m_xyForcedPlotRange.getXRange( );
                            }
                            else
                            {
                                xPosCorrected = m_xyForcedPlotRange.getMaximumX( ) - xPosCorrected * m_xyForcedPlotRange.getXRange( );
                            }
                            
                            yPosCorrected = m_xyForcedPlotRange.getMaximumY( ) - yPosCorrected * m_xyForcedPlotRange.getYRange( );
                        }
                        else
                        {
                            if ( m_displayXReversed == false )
                            {
                                xPosCorrected = m_xyAutoPlotRange.getMinimumX( ) + xPosCorrected * m_xyAutoPlotRange.getXRange( );
                            }
                            else
                            {
                                xPosCorrected = m_xyAutoPlotRange.getMaximumX( ) - xPosCorrected * m_xyAutoPlotRange.getXRange( );
                            }
                            
                            yPosCorrected = m_xyAutoPlotRange.getMaximumY( ) - yPosCorrected * m_xyAutoPlotRange.getYRange( );
                        }
                    }
                }
            }
        }
        
        return xPosCorrected;
    }
    
    
    /**
     * Determines the selected values at the position indicated by the
     * mouse event.
     */
    private void determineValuesAtCursorPosition( int xPos, int yPos )
    {
        determineValuesAtCursorPosition( new Point( xPos, yPos ) );
    }
    
    /**
     * Determines the selected values at the position indicated by the
     * mouse event.
     */
    protected Point2D.Double determineValuesAtCursorPosition( Point p )
    {
        
        MRSignalProperties mrcond = MRSignalProperties.getInstance( );
        
        if ( ( p.x > this.getSize( ).getWidth( ) / 10 ) && ( p.x < 9 * this.getSize( ).getWidth( ) / 10 ) )
        {   
            if ( ( p.y > this.getSize( ).getHeight( ) / 10 ) && ( p.y < 9 * this.getSize( ).getHeight( ) / 10 ) )
            {
                float xrange = 8.0f * (float)getSize( ).getWidth( ) / 10.0f;
                float yrange = 8.0f * (float)getSize( ).getHeight( ) / 10.0f;
                
                if ( m_curves != null )
                {
                    if ( m_curves.size( ) > 0 )
                    {
                        float xPosCorrected = ( ( (float)p.x - (float)this.getSize( ).getWidth( ) / 10.0f )  )  / xrange;
                        float yPosCorrected = ( ( (float)p.y - (float)this.getSize( ).getHeight( ) / 10.0f )  )  / yrange;
                        
                        if ( m_XYPlotRangeSelectorMenuItem.isSelected( ) == true )
                        {
                            if ( this.m_displayXReversed == true )
                            {
                                m_lastSelectedOffsetValue = m_xyForcedPlotRange.getMinimumX( ) + xPosCorrected * m_xyForcedPlotRange.getXRange( );
                            }
                            else
                            {
                                m_lastSelectedOffsetValue = m_xyForcedPlotRange.getMaximumX( ) + xPosCorrected * m_xyForcedPlotRange.getXRange( );
                            }
                            
                            m_lastSelectedAmplitudeValue = m_xyForcedPlotRange.getMaximumY( ) - yPosCorrected * m_xyForcedPlotRange.getYRange( );
                        }
                        else
                        {
                            if ( this.m_displayXReversed == false )
                            {
                                m_lastSelectedOffsetValue = m_xyAutoPlotRange.getMinimumX( ) + xPosCorrected * m_xyAutoPlotRange.getXRange( );                 
                            }
                            else
                            {   
                                m_lastSelectedOffsetValue = m_xyAutoPlotRange.getMaximumX( ) - xPosCorrected * m_xyAutoPlotRange.getXRange( );
                            }
                            m_lastSelectedAmplitudeValue = m_xyAutoPlotRange.getMaximumY( ) - yPosCorrected * m_xyAutoPlotRange.getYRange( );
                        }
                    }
                }
            }
        }
        
        if ( m_selectXYPointTriggerSet == true )
        {
            m_lastSelectedOffsetValue = convertToTDFDFitPointUnits( m_lastSelectedOffsetValue );
            m_lastSelectedAmplitudeValue = m_lastSelectedAmplitudeValue / ( 0.13975f * MRSignalProperties.getInstance( ).getNumberOfSamplePoints( ) );
        }
        
        return new Point2D.Double( (float)m_lastSelectedOffsetValue, (float)m_lastSelectedAmplitudeValue  );
    }
    
    /**
     * Converts the x-values of the user selected point to point units.
     * @param lastSelectedOffsetValue
     * @return converted offset value
     */
    public float convertToTDFDFitPointUnits( float lastSelectedOffsetValue )
    {
        MRSignalProperties mrcond = MRSignalProperties.getInstance( );
        float bandwidthHz = mrcond.getAcquistitionBandwidth( );
        int   samplePoints = mrcond.getNumberOfSamplePoints( );
        long  resonanceFrequency = mrcond.getResonanceFrequency( );
        float ppmAtCenter = mrcond.getPPMValueInCenterOfSpectrum( );
        
        float HzPerPPM = (float)(resonanceFrequency/1.0E6);
        float ppmBandwidth = bandwidthHz / HzPerPPM;
        float pointsPerPPM =  ppmBandwidth / samplePoints;
        
        TDFDFittingDefaults tdfdFitDefaults = TDFDFittingDefaults.getInstance( );
        tdfdFitDefaults.load( );
        
        float selectedPointNumber = -1.0f;
        
        
        if ( tdfdFitDefaults.getProperty( TDFDFittingDefaults.DISPLAY_UNIT ).compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 )
        {
            // nothing to do, just return the value
            selectedPointNumber = lastSelectedOffsetValue;
        }
        
        if ( tdfdFitDefaults.getProperty( TDFDFittingDefaults.DISPLAY_UNIT ).compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 )
        {
            selectedPointNumber = lastSelectedOffsetValue * samplePoints / bandwidthHz;
        }
        
        if ( tdfdFitDefaults.getProperty( TDFDFittingDefaults.DISPLAY_UNIT ).compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 )
        {
            
            float alfa = - ppmBandwidth/samplePoints;
            float beta = ( ppmAtCenter - alfa * 0.5f * samplePoints );
            
            selectedPointNumber = -samplePoints/2f + ( lastSelectedOffsetValue - beta ) / alfa;
        }
        
        return selectedPointNumber;
    }
        
    /**
     * Sets the measured reference spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setMeasuredSpectrum( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing )
    {
        m_measuredSpectrum = new SingleVoxelMeasuredSpectralData( );
        m_measuredSpectrum.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                    specData.getDispersionChannelTD( ) );
        
        m_measuredSpectrum.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_measuredSpectrum.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_measuredSpectrum.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_measuredSpectrum.setCurveColor( Color.black );
        addSpectralDataSet( "measured", m_measuredSpectrum, true );
                
        changeDisplayedSignal( );
    }
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setModelSpectrum( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing )
    {
        m_modelSpectrum = new SingleVoxelMeasuredSpectralData( );
        m_modelSpectrum.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                 specData.getDispersionChannelTD( ) );
        
        m_modelSpectrum.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_modelSpectrum.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_modelSpectrum.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
                     
        m_modelSpectrum.setCurveColor( Color.red );
        
        addSpectralDataSet( "model", m_modelSpectrum, false );
                
        changeDisplayedSignal( );
    }
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setSimulatedSpectrum( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_modelSpectrum = new SingleVoxelMeasuredSpectralData( );
        m_modelSpectrum.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                 specData.getDispersionChannelTD( ) );
        
        m_modelSpectrum.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_modelSpectrum.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_modelSpectrum.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_modelSpectrum.setCurveColor( Color.green );
        
        addSpectralDataSet( "model", m_modelSpectrum, false );
        if ( refresh == true )
        {        
            changeDisplayedSignal( );
        }
    }
    
    /**
     * Sets the plain resulting TDFDFit spectrum object in the m_modelSpectrumViewer
     * Not method call only from RandonForestRegression 
     * @param specData
     * @param clearBeforeDrawing
     * @param refresh
     */
    public void setPlainTDFDFittedSpectrum( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_modelSpectrumTDFDFitDefaultModel = new SingleVoxelMeasuredSpectralData( );
        m_modelSpectrumTDFDFitDefaultModel.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                                    specData.getDispersionChannelTD( ) );
        
        m_modelSpectrumTDFDFitDefaultModel.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_modelSpectrumTDFDFitDefaultModel.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_modelSpectrumTDFDFitDefaultModel.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_modelSpectrumTDFDFitDefaultModel.setCurveColor( Color.DARK_GRAY );
        
        addSpectralDataSet( "tdfdfitted-of", m_modelSpectrumTDFDFitDefaultModel, clearBeforeDrawing );
        if ( refresh == true )
        {
            changeDisplayedSignal( );
        }
    }
    
     /**
     * Sets the plain resulting TDFDFit spectrum object in the m_modelSpectrumViewer
     * Not method call only from RandonForestRegression 
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setRFImprovedTDFDFittedSpectrum( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_modelSpectrumTDFDFitRFImprovedModel = new SingleVoxelMeasuredSpectralData( );
        m_modelSpectrumTDFDFitRFImprovedModel.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                                       specData.getDispersionChannelTD( ) );
        
        m_modelSpectrumTDFDFitRFImprovedModel.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_modelSpectrumTDFDFitRFImprovedModel.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_modelSpectrumTDFDFitRFImprovedModel.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_modelSpectrumTDFDFitRFImprovedModel.setCurveColor( Color.CYAN );
        
        addSpectralDataSet( "tdfdfitted-rf-impr", m_modelSpectrumTDFDFitRFImprovedModel, clearBeforeDrawing );
        if ( refresh == true )
        {
            changeDisplayedSignal( );
        }
    }
    
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setPredictedSpectralModelResponse( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_predictedSpectrum = new SingleVoxelMeasuredSpectralData( );
        m_predictedSpectrum.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                     specData.getDispersionChannelTD( ) );
        
        m_predictedSpectrum.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_predictedSpectrum.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_predictedSpectrum.autoSetXaxis( );
        m_predictedSpectrum.rotateInFrequencyDomain(-1);
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_predictedSpectrum.setCurveColor( Color.ORANGE.darker( ) );
        
        addSpectralDataSet( "predicted model", m_predictedSpectrum, false );
        if ( refresh == true )
        {
            changeDisplayedSignal( );
        }
    }
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setGroundtruthModelResponse( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing )
    {
        m_groundTruthModelSpectrum = new SingleVoxelMeasuredSpectralData( );
        m_groundTruthModelSpectrum.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                            specData.getDispersionChannelTD( ) );
        
        m_groundTruthModelSpectrum.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_groundTruthModelSpectrum.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_groundTruthModelSpectrum.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_groundTruthModelSpectrum.setCurveColor( Color.MAGENTA );
        
        addSpectralDataSet( "groundtruth response", m_groundTruthModelSpectrum, false );
                
        changeDisplayedSignal( );
    }
    
     /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setNoiselessGroundTruthModelResponse( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_noiselessGroundTruthModelSpectrum = new SingleVoxelMeasuredSpectralData( );
        m_noiselessGroundTruthModelSpectrum.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                                     specData.getDispersionChannelTD( ) );
        
        m_noiselessGroundTruthModelSpectrum.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_noiselessGroundTruthModelSpectrum.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_noiselessGroundTruthModelSpectrum.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_noiselessGroundTruthModelSpectrum.setCurveColor( Color.BLUE );
        
        addSpectralDataSet( "noiseless groundtruth response", m_noiselessGroundTruthModelSpectrum, false );
        if ( refresh == true )
        {
            changeDisplayedSignal( );
        }
    }
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setNoisyDifferenceResponse( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_noisyDifferenceResponse = new SingleVoxelMeasuredSpectralData( );
        m_noisyDifferenceResponse.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                           specData.getDispersionChannelTD( ) );
        
        m_noisyDifferenceResponse.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_noisyDifferenceResponse.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_noisyDifferenceResponse.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_noisyDifferenceResponse.setCurveColor( Color.ORANGE );
        
        addSpectralDataSet( "noisy difference response", m_noisyDifferenceResponse, false );
        if ( refresh == true ) 
        {
            changeDisplayedSignal( );
        }
    }
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     */
    public void setNoisyGTDifferenceResponse( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_noisyGTDifferenceResponse = new SingleVoxelMeasuredSpectralData( );
        m_noisyGTDifferenceResponse.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                             specData.getDispersionChannelTD( ) );
        
        m_noisyGTDifferenceResponse.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_noisyGTDifferenceResponse.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_noisyGTDifferenceResponse.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_noisyGTDifferenceResponse.setCurveColor( Color.ORANGE );
        
        addSpectralDataSet( "noisy GT difference response", m_noisyGTDifferenceResponse, false );
        if ( refresh == true )
        {
            changeDisplayedSignal( );
        }
    }
        
    
    /**
     * Sets the model spectrum object in the m_modelSpectrumViewer
     * The reference spectrum, is the spectrum that is measured in order to obtain
     * prior knowledge information.
     * @param specData
     * @param clearBeforeDrawing
     * @param refresh
     */
    public void setNoiseLessDifferenceResponse( SingleVoxelMeasuredSpectralData specData, boolean clearBeforeDrawing, boolean refresh )
    {
        m_noisyLessDifferenceResponse = new SingleVoxelMeasuredSpectralData( );
        m_noisyLessDifferenceResponse.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ),
                                                               specData.getDispersionChannelTD( ) );
        
        m_noisyLessDifferenceResponse.setDicomHeaderInfo( specData.getDicomHeaderInfo( ) );
        m_noisyLessDifferenceResponse.setSiemensSpecificDicomHeaderInfo( specData.getSiemensSpecificDicomHeaderInfo( ) );
        m_noisyLessDifferenceResponse.autoSetXaxis( );
        
        specData.autoSetXaxis( );
        if ( clearBeforeDrawing == true )
        {
            m_responseList.clear( );
            m_responseNameList.clear( );
            m_responseListBuffer.clear( );
            deleteAllCurves( );
        }
        
        m_noisyLessDifferenceResponse.setCurveColor( Color.GRAY );
        
        addSpectralDataSet( "noiseless difference response", m_noisyLessDifferenceResponse, false );
        if ( refresh == true )
        {
            changeDisplayedSignal( );
        }
    }
    
    
    /**
     * All registered objects are notified that an event of type
     * actionEventType happened.
     * @param actionEventType
     * @param actionCommand
     */
    protected void notifyActionListeners( int actionEventType, String actionCommand )
    {
        Iterator it = m_listeners.iterator( );
        while( it.hasNext( ) )
        {
            ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( this, actionEventType, actionCommand ) );
        }
    }
    
    /**
     * Returns the offset value that was last selected by the user.
     * @return 
     */
    public float getLastSelectedOffsetValue( )
    {
        return m_lastSelectedOffsetValue;
    }
    
     /**
     * Returns the offset value that was last selected by the user.
     * @return 
     */
    public float getLastSelectedOffsetValueOfLastMouseClickedEvent( )
    {
        return m_lastSelectedOffsetValueAtMouseClickEvent;
    }
    
    /**
     * Returns the offset value that was last selected by the user.
     * @return 
     */
    public float getLastSelectedAreaValueOfLastMouseClickedEvent( )
    {
        return m_lastSelectedAmplitudeValueAtMouseClickEvent;
    }
    
    /**
     * Returns the amplitude value that was last selected by the user.
     * @return 
     */
    public float getLastSelectedAmplitudeValue( )
    {
        return m_lastSelectedAmplitudeValue;
    }
    
    /**
     * Returns the measured spectrum data object.
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getMeasuredSpectrum( )
    {
        return m_measuredSpectrum;
    }
    
    /**
     * Prints the spectrum.
     * @param g
     * @param pf
     * @param pageIndex
     * @return 
     * @throws java.awt.print.PrinterException
     */
    @Override
    public int print( Graphics g, PageFormat pf, int pageIndex ) throws PrinterException 
    {
        if ( pageIndex > 0 )
        {
            return Printable.NO_SUCH_PAGE;
        }
        Graphics2D g2d = (Graphics2D)g;
        g2d.translate( pf.getImageableX( ), pf.getImageableY( ) );
                
        Rectangle componentBounds = this.getBounds(null);
        double scaleX = pf.getImageableWidth( ) / componentBounds.width;
        double scaleY = pf.getImageableHeight( ) / componentBounds.height;
        /*
        if (scaleX < 1 || scaleY < 1) 
        {
            
            if (scaleX < scaleY) 
            {
                scaleY = scaleX;
            } 
            else 
            {
                scaleX = scaleY;
            }
        }
        */
        g2d.scale( scaleX, scaleY );
        
        paintComponent( g2d );
        
        return Printable.PAGE_EXISTS;  
    }
    
    /**
     * Prints the currently displayed spectrum.
     */
    public void print( )
    {
        //--- Create a new PrinterJob object
        PrinterJob printJob = PrinterJob.getPrinterJob( );
        
        //--- Create a new book to add pages to
        Book book = new Book( );
        
        //--- Add the document page using a landscape page format
        PageFormat documentPageFormat = new PageFormat( );
        documentPageFormat.setOrientation( PageFormat.LANDSCAPE );
        book.append ( this, documentPageFormat );
        
        //--- Tell the printJob to use the Book as the Pageable object
        printJob.setPageable( book );
        
        //--- Show the print dialog box. If the user clicks the
        //--- print button, we then proceed to print, else we cancel
        //--- the process.
        try 
        {
            if ( printJob.printDialog( ) )
            {
                printJob.print( );
            }
        }
        catch ( PrinterException e )
        {
        }  
    }
    
    
    // Attribututes of the class 
    private final LinkedList<ActionListener> m_listeners = new LinkedList<>( );
    
    private SingleVoxelMeasuredSpectralData m_measuredSpectrum;
    private SingleVoxelMeasuredSpectralData m_modelSpectrum;
    
    private SingleVoxelMeasuredSpectralData m_modelSpectrumTDFDFitDefaultModel;
    private SingleVoxelMeasuredSpectralData m_modelSpectrumTDFDFitRFImprovedModel;
    
    private SingleVoxelMeasuredSpectralData m_predictedSpectrum;
    private SingleVoxelMeasuredSpectralData m_groundTruthModelSpectrum;
    
    private SingleVoxelMeasuredSpectralData m_noiselessGroundTruthModelSpectrum;
    private SingleVoxelMeasuredSpectralData m_noisyDifferenceResponse;
    
    private SingleVoxelMeasuredSpectralData m_noisyGTDifferenceResponse; // this response should be pure noise
    private SingleVoxelMeasuredSpectralData m_noisyLessDifferenceResponse;
    
    
    // This menu item is the entry point allowing the user to edit the model
    // spectrum
    protected JMenu m_modelSpectumEditing;
    protected JMenuItem m_addLineMenuItem;
    protected JMenuItem m_printSpectrum;
    
    protected JMenuItem m_applyLineShapeTransform;
    
    // two coordidates allowing derived classes to obtain information where
    // mouse cursor popup.
    protected boolean m_selectXYPointTriggerSet = false;
    
    protected int m_mouseButtonClickedX;
    protected int m_mouseButtonClickedY;
    
    protected int m_mouseButtonPressedX;
    protected int m_mouseButtonPressedY;
    
    protected int m_mouseButtonReleasedX;
    protected int m_mouseButtonReleasedY;
    
    protected int m_mouseButtonDraggedX;
    protected int m_mouseButtonDraggedY;
    
    protected float m_lastSelectedOffsetValueAtMouseClickEvent;
    protected float m_lastSelectedAmplitudeValueAtMouseClickEvent;
    
    
    //protected float m_lastSelectedOffsetValue;
    //protected float m_lastSelectedAmplitudeValue;
    
    
    // event types that can be raised by this object
    public final static int ADD_LINE_EVENT = 0;
    public final static int APPLY_QUALITY_WATER_REFERENCE_PROCESSING_EVENT = 1;
    public final static int DO_NOT_APPLY_WATER_REFERENCE_PROCESSING_EVENT = 2;
    
    
    
    // Events related to use in a MultiXYPlotView
    public final static int    USER_REQUESTED_XYPLOT_ZOOM_EVENT = 100;
    public final static String USER_REQUESTED_XYPLOT_ZOOM_EVENT_CMD = "USER_REQUESTED_XYPLOT_ZOOM_EVENT_CMD";
    boolean m_mouseDragged = false;
    
    
    private TwoCompartmentWaterReferenceModelFittingFrame m_waterReferenceFrame;
    
    // For proper serialization
    private static final long serialVersionUID = 1486625898865493L;

   

}
