/*
 * File: $Id: RandomMinimizer.java,v 1.2 2006/04/20 09:29:29 hans Exp $
 * History
 * Date: 18.04.2006, 10:58  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2006 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.tdfdfit;

import org.spectrim.lib.DoubleRange;
import org.spectrim.xyplot.Curve;
import org.spectrim.xyplot.XYPlotFrame;

import java.util.Random;

/**
 * Modula-2 comment:
 * (**************************************************************************)
 * (*** This program minimizes an P dimensional function                   ***)
 * (*** J(p1,p2,..) in a random way, i.e. by most likehood distributions   ***)
 * (*** Test version 1.0 by H. Slotboom / april 30th, 1993                 ***)
 * (**************************************************************************)
 *
 * Recoded in Java by hs on 18-04-2006
 *
 * @author  hans
 */

public abstract class RandomMinimizer 
{
        
    /** 
     * Creates a new instance of RandomMinimizer 
     */
    public RandomMinimizer( )
    {
        m_n_iter_max = 100;
        m_number_of_random_trials = 100;
    }
 
    /** 
     * Computes the function values of the function to minimize
     */
     public abstract double fnc( double[] par );
    
    /**
     * Initializes the parameter vector
     */
    public void setStartParameters( double[] par )
    {
        
        m_par = new double[par.length];
        
        for ( int i = 0 ; i < par.length ; i++ )
        {
            m_par[i] = par[i];
        }    
    }
  
    /**
     * Initializes the parameter ranges vector
     */
    public void setParameterRanges( DoubleRange[] par_range )
    {
        m_parameterRanges = new DoubleRange[ par_range.length ];
        
        for ( int i = 0 ; i < par_range.length ; i++ )
        {
            DoubleRange parRange = new DoubleRange( par_range[i].getMinimum( ), par_range[i].getMaximum( ) );
            m_parameterRanges[i] = parRange;
        }
    }
    
    /**
     * Checks the starting value parameters on their validity
     */
    private boolean CheckParameters( )
    {
        
        boolean returnFlag = true;
        
        for ( int i = 0; i < m_par.length ; i++ )
        {
            if ( ( m_par[i] > m_parameterRanges[i].getMinimum( ) ) && ( m_par[i] < m_parameterRanges[i].getMaximum( ) ) )
            {
                ; // everything is ok
            }
            else
            {
                returnFlag = false;
            }  
        }
        
        if ( returnFlag == true ) 
        {
            if ( m_verbose == true )
            {
                System.out.println( "All initial parameters are O.K. " );
                System.out.println( "" );
            }
        }
        else
        {
            System.out.println(" *** > Corrupt initial parameter array in class SCANalyze.SignalProcessing.RandomMinimizer: ");
            System.out.println(" Initial parameter values do not satisfy the parameter boundary conditions ");
        }
        
        return returnFlag;
    }

    
    /**
     * Initializes the random search algorithm
     */
    private void InitRandomSearch( )
    {
        
        double ran_fval;
        
        for ( int j = 0 ; j < m_par.length; j++ )
        {
            m_par[j] = 0.5 * ( m_parameterRanges[j].getMaximum( ) + m_parameterRanges[j].getMinimum( ) ) + ( m_parameterRanges[j].getMaximum( ) - m_parameterRanges[j].getMinimum( ) ) * ( rnd.nextDouble( ) - 0.5 );
        }
    
        m_tmp_min_fnc = fnc( m_par );
        
        for ( int i = 0; i < m_par.length ; i++ )
        {
            m_parMinimum[i] = m_par[i];
        }
        
        for ( int i = 0 ; i < m_number_of_random_trials - 1 ; i++ )
        {
            
            for( int j = 0 ; j < m_par.length; j++ )
            {
                m_par[j] = 0.5 * ( m_parameterRanges[j].getMaximum( ) + m_parameterRanges[j].getMinimum( ) ) + ( m_parameterRanges[j].getMaximum( ) - m_parameterRanges[j].getMinimum( ) ) * ( rnd.nextDouble( ) - 0.5 );
            }
            
            ran_fval = fnc( m_par );
            
            if ( ran_fval < m_tmp_min_fnc )
            {
                m_tmp_min_fnc = ran_fval;
                
                m_parMinimum = copyVector( m_par );
            }
            
        }
    }

    /**
     * Directs the values of a double vector to standard output
     */
    public void displayVector( double[] invec )
    {
        for( int i = 0 ; i < invec.length; i++ )
        {
            System.out.println( "Par[" + i + "] = " + invec[i] );
        }
    }

    /**
     * Returns a copy of double precision inpar arrqay
     */
    private double[] copyVector( double[] inpar )
    {   
        
        double[] outpar = new double[ inpar.length];
        
        for ( int i = 0 ; i < inpar.length; i++ )
        {
            outpar[i] = inpar[i];
        }    
        return outpar;
    }
    
    
    /**
     * Computes a random parameter vector
     */
    private double[] RandomPar( int k, double[] center )
    {
        double ranval, ran1, ran2;
        
        double[] tmpPar = new double[center.length];
        
        for ( int i = 0; i < center.length; i++ )
        {
            double tstMin = m_parameterRanges[i].getMinimum( );
            double tstMax = m_parameterRanges[i].getMaximum( );

            do
            {
                ran1 = 2.0 * ( rnd.nextDouble( ) - 0.5 );
                ran2 = rnd.nextDouble( ) - 0.5;
                if( ran2 != 0.0 )
                {
                    ran2 = ran2 / Math.abs( ran2 ) ;
                }
                else
                {
                    ran2 = 1.0;
                }
                
                ranval = center[i] + ( m_parameterRanges[i].getMaximum( ) - m_parameterRanges[i].getMinimum( ) ) * ran2 * Math.exp( -(float)k * Math.abs( ran1 ) );
            }
            while ( ( ranval < tstMin ) || ( ranval > tstMax ) );
      
            // System.out.print( '*' );
        
            tmpPar[i] = ranval;
        }
        
        return tmpPar;
        
    }
    
    
    /**
     * Performs a random minimization
     */
    private void randomMinimize( )
    {
        
        double  ran_fval;
        
        double[] convmin = new double[m_n_iter_max + 1]; 
        
        double[] xaxis   = new double[m_n_iter_max + 1];

        m_tmp_min_fnc = fnc( m_parMinimum );
        
        xaxis[0] = 0.0;
        convmin[0] = m_tmp_min_fnc;
        
        double[] tmp_par = new double[m_par.length];
        double[] tmp_par_min = new double[m_par.length];
        
        
        tmp_par_min = copyVector( m_parMinimum );
        
        
        for ( int k = 1 ; k <= m_n_iter_max ; k++ )
        {
            /* search for a new minimum */
            for ( int i = 0 ; i < m_number_of_random_trials; i++ )
            {
                tmp_par = RandomPar( k, m_parMinimum );
                ran_fval = fnc( tmp_par );
                if ( ran_fval < m_tmp_min_fnc )
                {
                    m_tmp_min_fnc = ran_fval;
                    tmp_par_min = copyVector( tmp_par );
                }
            }
            
            m_parMinimum = copyVector( tmp_par_min );
            
            if ( m_verbose == true )
            {
                System.out.println( "" );
                System.out.println( "After the " + k + "-th iteration : " );
                System.out.println( "" );
                System.out.println( "  Vector containing minimizing parameters : " );
                displayVector( m_parMinimum );
                System.out.println( "  Minimum function-value : " + m_tmp_min_fnc );
                System.out.println("");
            }

            convmin[k] = m_tmp_min_fnc;
            
            xaxis[k] = (float)k;
            
        }
        
        if ( convmin[0] != 0.0 )
        {
            double factor = convmin[0];
            for ( int i=0 ; i < convmin.length; i++ )
            {
                convmin[i] = convmin[i]/factor;
            }
        }
        if ( m_verbose == true )
        {
            XYPlotFrame convMinPlot = new XYPlotFrame( "Minimum function values", null );
            convMinPlot.addCurve( new Curve( xaxis, convmin ), true );
        }

        if ( m_verbose == true )
        {
            System.out.println( "Obtained minimum : " );
            System.out.println( "" );
            System.out.println( "" );
        
            System.out.println( "  Vector containing minimizing parameters : " );
            System.out.println( "");
            displayVector( m_parMinimum );
            System.out.println("  Minimum function-value : " + m_tmp_min_fnc );
            System.out.println( "" );
        }

    }
    
    /**¨
     * Returs the sum of least squares
     * @return
     */
    public double getMinimumFucntionValue( )
    {
        return m_tmp_min_fnc;
    }


    /**
     *  Performs the random minimization
     *  This minimization/maximization algorithm minimizes and maximizes  
     *  a given p-dimensional function f(par) on a given limited finite   
     *  parameter domain. The minimization is carried out randomly. The   
     *  user has to specify the maximum number of iterations n_itermax,   
     *  the number of random function evaluations per iteration N, the    
     *  parameter domain 'par_range' in which the minimum and maximum of  
     *  the function has to be found and whether an initial parameter-    
     *  vector is available. In the case that init_val_avail is FALSE the 
     *  algorithm will produce its own initial guess.
     */
    public boolean performRandomMinimization( )
    {
        boolean flag = true;
        
        if ( ( m_par.length > 0 ) && ( m_par.length == m_parameterRanges.length ) )
        {    
            if ( CheckParameters( ) == true )
            {
                
                m_parMinimum = copyVector( m_par );
                
                
                InitRandomSearch( );    
                
                randomMinimize( );
            }
            else
            {
                flag = false;
                System.out.println( " Failure : Check initial parameter values and parameter range. ");
                System.out.println( " And try again ");
            }
        }
        else
        {
            System.out.println( "Initial Parameter vector and/or parameter ranges vector missing or incomplete ");
        }
        
        return flag;
        
    }
    
    
    /**
     * Performs the random minimization
     */
    public boolean performRandomMinimization( int numberOfRandomTrials, int niter_max, boolean verbose )
    {
        boolean flag = true;
        m_verbose = verbose;
        m_number_of_random_trials = numberOfRandomTrials;
        m_n_iter_max = niter_max;
        //  This minimization/maximization algorithm minimizes and maximizes  
        //  a given p-dimensional function f(par) on a given limited finite   
        //  parameter domain. The minimization is carried out randomly. The   
        //  user has to specify the maximum number of iterations n_itermax,   
        //  the number of random function evaluations per iteration N, the    
        //  parameter domain 'par_range' in which the minimum and maximum of  
        //  the function has to be found and whether an initial parameter-    
        //  vector is available. In the case that init_val_avail is FALSE the 
        //  algorithm will produce its own initial guess.                     
        //

        if ( ( m_par.length > 0 ) && ( m_par.length == m_parameterRanges.length ) )
        {    
            if ( CheckParameters( ) == true )
            {
                
                m_parMinimum = copyVector( m_par );
                
                
                InitRandomSearch( );    
                
                randomMinimize( );
            }
            else
            {
                flag = false;
                System.out.println( " Failure : Check initial parameter values and parameter range. ");
                System.out.println( " And try again ");
            }
        }
        else
        {
            System.out.println( "Initial Parameter vector and/or paramter ranges vector missing or incomplete ");
        }
        
        return flag;
        
    }
    
    /**
     * Returs the value of the parameter vector m_parMinimum
     * which contains the optimal parameters
     */
    public double[] getMinimizingParameters( )
    {
        return m_parMinimum;
    }
    
    /**
     * Sets the number of maximum allowed iterations
     */
    public void setMaxNumberOfIterations( int maxIter )
    {
        this.m_n_iter_max = maxIter;
    }
            
    /**
     * Sets the number of maximum allowed iterations
     */
    public int getMaxNumberOfIterations( )
    {
        return m_n_iter_max;
    }        
            
    /**
     * Sets the number of maximum allowed iterations
     */
    public void setMaxNumberOfRandomTries( int maxRandomTrials )
    {
        m_number_of_random_trials = maxRandomTrials;
    }
            
    /**
     * Sets the number of maximum allowed iterations
     */
    public int setMaxNumberOfRandomTries( )
    {
        return m_number_of_random_trials;
    }       
    
    // Member variables
    
    private double[] m_par;
    private DoubleRange[] m_parameterRanges;
    private Random rnd = new Random( );
    
    private int m_n_iter_max;
    private int m_number_of_random_trials;
    private double m_tmp_min_fnc;
   
    
    private double[] m_parMinimum;
    private boolean m_verbose = true;
   
}