/*
 * File: $Id: SingleVoxelSpectralData.java,v 1.2 2006/04/26 14:54:11 hans Exp $
 * History
 * Date: 09.05.2003, 10:58  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */
  
package org.spectrim.tdfdfit;

import org.spectrim.signalprocessing.FFT;
import java.awt.Color;

import java.io.RandomAccessFile;
import java.io.IOException;

import java.nio.ByteBuffer;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import java.util.LinkedHashMap;
import java.util.Iterator;


import org.spectrim.dicom.FileInfo;
import org.spectrim.imagegraphicobjects.Spectroscopic3DVoxel;
import org.spectrim.imagegraphicobjects.Point3D;
import org.spectrim.xyplot.Curve;


/**
 * Container class containing the data of a one dimensional spectrum. 
 * @author  hans
 */
public class SingleVoxelSpectralData 
{
    
    /** 
     * Creates a new instance of SingleVoxelData. 
     */
    public SingleVoxelSpectralData( )
    {
        
    }
        
    /** 
     * Creates a new instance of SingleVoxelData (copy constructor).
     * @param svsData
     */
    public SingleVoxelSpectralData( SingleVoxelSpectralData svsData )
    {
        // Attributes :
        m_numberOfSamplePoints = svsData.m_numberOfSamplePoints;  
        
        // Both channel (alterating abs, disp) dispersion and absorption channel
        // in time domain (TD)
        if ( svsData.m_bothChannelsTD != null )
        {
            m_bothChannelsTD = new float[svsData.m_bothChannelsTD.length];
            System.arraycopy( svsData.m_bothChannelsTD, 0, m_bothChannelsTD, 0, svsData.m_bothChannelsTD.length ); 
        }
        
        if ( svsData.m_absorptionChannelTD != null )
        {
            m_absorptionChannelTD = new float[ svsData.m_absorptionChannelTD.length ];
            System.arraycopy( svsData.m_absorptionChannelTD, 0, m_absorptionChannelTD, 0, svsData.m_absorptionChannelTD.length ); 
        }
        
        if ( svsData.m_dispersionChannelTD != null )
        {
            m_dispersionChannelTD = new float[ svsData.m_dispersionChannelTD.length ];
            System.arraycopy( svsData.m_dispersionChannelTD, 0, m_dispersionChannelTD, 0, svsData.m_dispersionChannelTD.length ); 
        }
        
        if ( svsData.m_xaxisTD != null )
        {
            m_xaxisTD = new float[svsData.m_xaxisTD.length ];
            System.arraycopy( svsData.m_xaxisTD, 0, m_xaxisTD, 0, svsData.m_xaxisTD.length );
        }
        
        if ( svsData.m_bothChannelsFD != null )
        {
            m_bothChannelsFD = new float[ svsData.m_bothChannelsFD.length ];
            System.arraycopy( svsData.m_bothChannelsFD, 0, m_bothChannelsFD, 0, svsData.m_bothChannelsFD.length );
        }
        
        if ( svsData.m_absorptionChannelFD != null )
        {
            m_absorptionChannelFD = new float[ svsData.m_absorptionChannelFD.length ];
            System.arraycopy( svsData.m_absorptionChannelFD, 0, m_absorptionChannelFD, 0, svsData.m_absorptionChannelFD.length );
        }
        
        if ( svsData.m_dispersionChannelFD != null )
        {
            m_dispersionChannelFD = new float[svsData.m_dispersionChannelFD.length];
            System.arraycopy( svsData.m_dispersionChannelFD, 0, m_dispersionChannelFD, 0, svsData.m_dispersionChannelFD.length );
        }
        
        if ( svsData.m_xaxisFD != null )
        {
            m_xaxisFD = new float[svsData.m_xaxisFD.length];
            System.arraycopy( svsData.m_xaxisFD, 0, m_xaxisFD, 0, svsData.m_xaxisFD.length );
        }
        
        m_dicomHeaderInfo = svsData.m_dicomHeaderInfo;
        m_curveColor = svsData.m_curveColor;
        m_curveLabel = svsData.m_curveLabel;

        m_siemensHeaderInfo = svsData.m_siemensHeaderInfo;
    
        m_referenceDicomFile0Path = svsData.m_referenceDicomFile0Path;
        m_referenceDicomFile1Path = svsData.m_referenceDicomFile1Path;
        m_referenceDicomFile2Path = svsData.m_referenceDicomFile2Path;
    
        m_voxelGeometry = svsData.m_voxelGeometry;
    
        m_sVOIThickness = svsData.m_sVOIThickness;
        m_sVoIPhaseFOV = svsData.m_sVoIPhaseFOV;
        m_sVoIReadoutFOV = svsData.m_sVoIReadoutFOV;
    
        m_dSagVoxelPosition = svsData.m_dSagVoxelPosition;
        m_dCorVoxelPosition = svsData.m_dCorVoxelPosition;
        m_dTraVoxelPosition = svsData.m_dTraVoxelPosition;
    
        m_dSagVoxelOrientation = svsData.m_dSagVoxelOrientation;
        m_dCorVoxelOrientation = svsData.m_dCorVoxelOrientation;
        m_dTraVoxelOrientation = svsData.m_dTraVoxelOrientation;
    
        m_dInPlaneRotation = svsData.m_dInPlaneRotation;
       
        MRSignalProperties signalProperties = MRSignalProperties.getInstance( );
        
        
        // Value necessary to compute exact ppm values
        m_resonanceFrequency = signalProperties.getResonanceFrequency( );
    
        // Value from Siemens specific header to compute sample frequency
        m_sampleDwellTime = svsData.m_sampleDwellTime;
        m_acqFrequency = signalProperties.getAcquistitionBandwidth( );
    }
    
    
    /** 
     * Creates a new instance of SingleVoxelData.
     * @param color
     */
    public SingleVoxelSpectralData( Color color )
    {
        m_curveColor = color;
    }
    
    /**
     * Creates a new instance of SingleVoxelData and reserves N sample
     * points for the response.
     * @param samplePoints
     */
    public SingleVoxelSpectralData( int samplePoints )
    {
       m_numberOfSamplePoints = samplePoints;
       allocateArrays( samplePoints );
    }
    
    /**
     * Creates a new instance of SingleVoxelData and reserves N sample
     * points for the response.
     * @param samplePoints
     * @param color
     */
    public SingleVoxelSpectralData( int samplePoints, Color color )
    {
       m_curveColor = color;
       m_numberOfSamplePoints = samplePoints;
       allocateArrays( m_numberOfSamplePoints );
    }

    /**
     * Creates a new instance of SingleVoxelData and reserves N sample
     * points for the response.
     * @param abs
     * @param disp
     * @param domain
     */
    public SingleVoxelSpectralData( float[] abs, float[] disp, int domain )
    {
        
        m_numberOfSamplePoints = abs.length;
        
        if ( domain == SingleVoxelSpectralData.TIME_DOMAIN )
        {
            m_curveColor = Color.BLACK;
            allocateArrays( abs.length );
            for ( int i=0 ; i < abs.length ; i++ )
            {
                this.m_absorptionChannelTD[i]=abs[i];
                this.m_dispersionChannelTD[i]=disp[i];
                m_bothChannelsTD[ 2 * i ] = abs[i];
                m_bothChannelsTD[ 2 * i + 1] = disp[i];
            }
       
            // brings the frequency domain response into sync
            calculateFDRepresentation( );
        }
        else
        {
            m_curveColor = Color.BLACK;
            allocateArrays( abs.length );
            for ( int i=0 ; i < abs.length ; i++ )
            {
                this.m_absorptionChannelFD[i]=abs[i];
                this.m_dispersionChannelFD[i]=disp[i];
                m_bothChannelsFD[ 2 * i ] = abs[i];
                m_bothChannelsFD[ 2 * i + 1] = disp[i];
            }
       
            // brings the frequency domain response into sync
            calculateTDRepresentation( );
        }
    }
    
    /**
     * 
     * Returns the difference in minimum and maximum absorption spectrum value.
     * 
     * @return 
     */
    public float getDynamicSpectrumRangeAbsorption( )
    {
        float maxSpecVal = SignalTools.getMaximum( m_absorptionChannelFD );
        float minSpecVal = SignalTools.getMinimum( m_absorptionChannelFD );
        return (maxSpecVal - minSpecVal);
    }
    
    /**
     * Subtract a constant value of both absorption and dispersion channel. 
     * @param measuredSpectrumAbsRange 
     */
    void subtractConstantValueFromBothFDChannels(float measuredSpectrumAbsRange) 
    {
        
        for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
        {
            m_absorptionChannelFD[i] = m_absorptionChannelFD[i] - measuredSpectrumAbsRange/5f;
            m_dispersionChannelFD[i] = m_dispersionChannelFD[i] - measuredSpectrumAbsRange/5f;
        }
    }
    
    /**
     * Performs signal addition of this with other SingleVoxelSpectralData-data 
     * object. 
     * @param svsData
     * @return 
     */
    public boolean addSingleVoxelSpectralDataObject (SingleVoxelSpectralData svsData )
    {
        if ( this.m_absorptionChannelTD.length != svsData.getAbsorptionChannelTD( ).length )
        {
            return false;
        }
        else
        {
            float[] absTD = new float[ m_absorptionChannelTD.length ];
            absTD = svsData.getAbsorptionChannelTD( );
            float[] dispTD = new float[ m_dispersionChannelTD.length ];
            dispTD = svsData.getDispersionChannelTD( );
            
            m_bothChannelsTD = new float[ 2 * m_dispersionChannelTD.length ];
            
            for ( int i = 0; i < absTD.length; i++ )
            {
                this.m_absorptionChannelTD[i] += absTD[i];
                this.m_dispersionChannelTD[i] += dispTD[i];
                this.m_bothChannelsTD[2*i] = this.m_absorptionChannelTD[i];
                this.m_bothChannelsTD[2*i + 1] = this.m_dispersionChannelTD[i];
            }
            // Brings the frequency domain response into sync
            calculateFDRepresentation( );
        }
        return true;
    }
    
    /**
     * Performs signal subtraction of this with other 
     * SingleVoxelSpectralData-data object.
     * @param svsData
     * @return 
     */
    public boolean subtractSingleVoxelSpectralDataObject (SingleVoxelSpectralData svsData )
    {
        if ( this.m_absorptionChannelTD.length != svsData.getAbsorptionChannelTD( ).length )
        {
            return false;
        }
        else
        {
            float[] absTD = new float[ m_absorptionChannelTD.length ];
            absTD = svsData.getAbsorptionChannelTD( );
            float[] dispTD = new float[ m_dispersionChannelTD.length ];
            dispTD = svsData.getDispersionChannelTD( );
            
            m_bothChannelsTD = new float[ 2 * m_dispersionChannelTD.length ];
            
            for ( int i = 0; i < absTD.length; i++ )
            {
                this.m_absorptionChannelTD[i] -= absTD[i];
                this.m_dispersionChannelTD[i] -= dispTD[i];
                this.m_bothChannelsTD[2*i] = this.m_absorptionChannelTD[i];
                this.m_bothChannelsTD[2*i + 1] = this.m_dispersionChannelTD[i];
            }
            // brings the frequency domain response into sync
            calculateFDRepresentation( );
        }
        return true;
    }
    
    
    /**
     * Allocates the working data arrays.
     * @param samplePoints
     */
    protected void allocateArrays( int samplePoints )
    {
        m_numberOfSamplePoints = samplePoints;
        m_bothChannelsTD = new float[2 * samplePoints];
        m_absorptionChannelTD = new float[samplePoints];
        m_dispersionChannelTD = new float[samplePoints];
        m_xaxisTD = new float[samplePoints];

        m_bothChannelsFD = new float[2 * samplePoints];
        m_absorptionChannelFD = new float[samplePoints];
        m_dispersionChannelFD = new float[samplePoints];
        m_xaxisFD = new float[samplePoints];
    }
    
    
    /**
     * Allocates the working data arrays.
     * @param samplePoints
     */
    public void reallocateArrays( int samplePoints )
    {
        m_numberOfSamplePoints = samplePoints;
        m_bothChannelsTD = new float[2 * samplePoints];
        m_absorptionChannelTD = new float[samplePoints];
        m_dispersionChannelTD = new float[samplePoints];
        m_xaxisTD = new float[samplePoints];

        m_bothChannelsFD = new float[2 * samplePoints];
        m_absorptionChannelFD = new float[samplePoints];
        m_dispersionChannelFD = new float[samplePoints];
        m_xaxisFD = new float[samplePoints];
    }
    
    /**
     * Adds a time response to the current response.
     * @param absorption
     * @param dispersion
     */
    public void addTimeDomainResponse( float[] absorption, float[] dispersion )
    {
        setNumberOfSamplePoints( absorption.length );
        // ensure that the datastructures are defined!
        if ( m_absorptionChannelTD == null )
        {
            m_absorptionChannelTD = new float[absorption.length];
            m_dispersionChannelTD = new float[dispersion.length];
        }
        
        if ( ( absorption.length == m_absorptionChannelTD.length ) &&
             ( dispersion.length == m_dispersionChannelTD.length ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length; i++ )
            {
                m_absorptionChannelTD[i] += absorption[i];
                m_dispersionChannelTD[i] += dispersion[i];
            }
        }
        
        // Brings the frequency domain response into sync with TD
        calculateFDRepresentation( );
    }
    
     /**
     * Adds a time response to the current response.
     * @param absorption
     * @param dispersion
     */
    public void setNewTimeDomainResponse( float[] absorption, float[] dispersion )
    {
        setNumberOfSamplePoints( absorption.length );
        // ensure that the datastructures are defined!
        if ( m_absorptionChannelTD == null )
        {
            m_absorptionChannelTD = new float[absorption.length];
            m_dispersionChannelTD = new float[dispersion.length];
        }
        
        if ( ( absorption.length == m_absorptionChannelTD.length ) &&
             ( dispersion.length == m_dispersionChannelTD.length ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length; i++ )
            {
                m_absorptionChannelTD[i] = absorption[i];
                m_dispersionChannelTD[i] = dispersion[i];
            }
        }
        
        // Brings the frequency domain response into sync with TD
        calculateFDRepresentation( );
    }
    
    
    /**
     * Adds a frequency response to the current response.
     * @param absorption
     * @param dispersion
     */
    public void addFrequencyDomainResponse( float[] absorption, float[] dispersion )
    {
        if ( ( absorption.length == m_absorptionChannelFD.length ) &&
             ( dispersion.length == m_dispersionChannelFD.length ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length; i++ )
            {
                m_absorptionChannelFD[i] += absorption[i];
                m_dispersionChannelFD[i] += dispersion[i];
            }
        }
        // brings the time domain response into sync.
        calculateTDRepresentation( );
    }
    
    /**
     * Subtract a time response to the current response.
     * @param absorption
     * @param dispersion
     */
    public void subtractTimeDomainResponse( float[] absorption, float[] dispersion )
    {
        if ( ( absorption.length == m_absorptionChannelTD.length ) &&
             ( dispersion.length == m_dispersionChannelTD.length ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length; i++ )
            {
                m_absorptionChannelTD[i] -= absorption[i];
                m_dispersionChannelTD[i] -= dispersion[i];
            }
        }
        // brings the frequency domain response into sync
        calculateFDRepresentation( );
    }
    
    /**
     * Subtract a frequency response to the current response.
     * @param absorption
     * @param dispersion
     */
    public void subtractFrequencyDomainResponse( float[] absorption, float[] dispersion )
    {
        if ( ( absorption.length == m_absorptionChannelFD.length ) &&
             ( dispersion.length == m_dispersionChannelFD.length ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length; i++ )
            {
                m_absorptionChannelFD[i] -= absorption[i];
                m_dispersionChannelFD[i] -= dispersion[i];
            }
        }
        // brings the time domain response into sync
        calculateTDRepresentation( );
    }
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels.
     */
     public void calculateFDRepresentation( )
    {
        float[][] fftArray = new float[ m_absorptionChannelTD.length][2];
        
        for ( int i = 0 ; i < m_absorptionChannelTD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelTD[i];
            fftArray[i][1] = m_dispersionChannelTD[i];
        }
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        m_absorptionChannelFD = new float[m_absorptionChannelTD.length];
        m_dispersionChannelFD = new float[m_dispersionChannelTD.length];

        for ( int i = 0 ; i < m_absorptionChannelTD.length ; i++ )
        {
            int d = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelTD.length;
            m_absorptionChannelFD[d] = fftArray[i][0];
            m_dispersionChannelFD[d] = fftArray[i][1];
        }
    
        fftArray = null;
    }
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels.
     */
    public void calculateTDRepresentation( )
    {
        float[][] fftArray = new float[m_absorptionChannelFD.length][2];
        float[] bufAbs = new float[m_absorptionChannelFD.length];
        float[] bufDisp = new float[m_absorptionChannelFD.length];
               
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelFD[ m_absorptionChannelFD.length - 1 - i ];
            fftArray[i][1] = m_dispersionChannelFD[ m_absorptionChannelFD.length - 1 - i ];
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
             bufAbs[ i ] = fftArray[i][0];
             bufDisp[ i ] = fftArray[i][1];           
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            int d = ( i + 1 + m_absorptionChannelFD.length / 2 ) % m_absorptionChannelFD.length;
            fftArray[d][0] = bufAbs[i];
            fftArray[d][1] = bufDisp[i];
        }
        
        bufAbs = null;
        bufDisp = null;
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        m_absorptionChannelTD = new float[m_absorptionChannelFD.length];
        m_dispersionChannelTD = new float[m_dispersionChannelFD.length];

        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            //int d = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelFD.length;
            m_absorptionChannelTD[i] = fftArray[i][0] / m_absorptionChannelTD.length;
            m_dispersionChannelTD[i] = fftArray[i][1] / m_absorptionChannelTD.length;
        }
    
        fftArray = null;
    }
    
    /**
     * Sets the default time and frequency domain x-axis.
     */
    public void autoSetXaxis( )
    {
        autoSetXaxis( m_numberOfSamplePoints );
    }
    
    /**
     * Sets the default time and frequency domain x-axis.
     * @param length
     */
    protected void autoSetXaxis( int length )
    {
        
        TDFDFittingDefaults fitdef = TDFDFittingDefaults.getInstance( );
        fitdef.load( );
        
        String displayDomain = fitdef.getProperty( TDFDFittingDefaults.DISPLAY_DOMAIN );
        String displayUnits = fitdef.getProperty( TDFDFittingDefaults.DISPLAY_UNIT );
        
        if ( displayDomain.compareTo( TDFDFittingDefaults.TIME_DOMAIN ) == 0 )
        {
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.TIME_DOMAIN_POINTS ) == 0 )
            {
                calculateTDXaxisPoints( length );
            }
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.TIME_DOMAIN_SECONDS ) == 0 )
            {
                calculateTDXaxisSeconds( length );
            }
        }

        if ( displayDomain.compareTo( TDFDFittingDefaults.FREQ_DOMAIN ) == 0 )
        {
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 )
            {
                calculateFDXaxisPoints( length );
            }
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 )
            {
                calculateFDXaxisHertz( length );
            }

            if ( displayUnits.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 )
            {
                calculateFDXaxisPPM( length );
            }
        }       
    }
    
    /**
     * Calculates the time domain x-axis in point units.
     * @param length
     */
    protected void calculateTDXaxisPoints( int length )
    {
        m_xaxisTD = new float[ length ];
        
        for ( int i= 0; i < length; i++ )
        {
            m_xaxisTD[i] = (float)i;
        }
    }

    /**
     * Calculates the frequency domain x-axis in point units.
     * @param length
     */
    protected void calculateFDXaxisPoints( int length )
    {
        m_xaxisFD = new float[ length ];
        
        for ( int i= 0; i < length; i++ )
        {
            m_xaxisFD[i] = (float)( i - ( length / 2 ) );
        }
    }
    
    /**
     * Calculates the frequency domain x-axis in "point" units.
     * @param length
     */
    protected void calculateTDXaxisSeconds( int length )
    {
        m_xaxisTD = new float[ length ];
        
        MRSignalProperties cond = MRSignalProperties.getInstance( );
        float specWidth = cond.getFrequencyScale( ) * length;
        
        for ( int i= 0; i < length; i++ )
        {
            m_xaxisTD[i] = (float)i / specWidth;
        }
    }

    /**
     * Calculates the frequency domain x-axis in Hertz units.
     * @param length
     */
    protected void calculateFDXaxisHertz( int length )
    {
        m_xaxisFD = new float[ length ];
        
        for ( int i = 0; i < length; i++ )
        {
            m_xaxisFD[i] = ( (float)( i - length / 2 ) / length ) * MRSignalProperties.getInstance( ).getAcquistitionBandwidth( );
        }
    }
    
    /**
     * Calculates the frequency domain x-axis in ppm units.
     * @param length
     */
    protected void calculateFDXaxisPPM( int length )
    {
        m_xaxisFD = new float[ length ];
              
        for ( int i = 0; i < length; i++ )
        {
            m_xaxisFD[i] = MRSignalProperties.getInstance( ).getPPMValueInCenterOfSpectrum( ) - ( (float)( i - length / 2 ) / length )  * MRSignalProperties.getInstance( ).getAcquistitionBandwidth( ) / ( MRSignalProperties.getInstance( ).getResonanceFrequency( ) / 1000000f );
        }
        
    }

    /** 
     * Set spectral data points.
     * @param bothChannelsTD
     */
    public void setSpectralDataPointsTD( float[] bothChannelsTD )
    {
        // bothChannels can directly be copied into m_bothChannels
        System.arraycopy( bothChannelsTD, 0, m_bothChannelsTD, 0, bothChannelsTD.length );
        
        m_numberOfSamplePoints = bothChannelsTD.length / 2;
        
        for ( int i = 0; i < bothChannelsTD.length / 2; i++ )
        {
            m_absorptionChannelTD[ i ] = bothChannelsTD[ 2 * i ];
            m_dispersionChannelTD[ i ] = bothChannelsTD[ 2 * i + 1 ];
        }
        
        calculateFDRepresentation( );
        
        // Set a default xaxis (time domain)
        autoSetXaxis( );
    }
    
    /** 
     * Set time domain spectral data points; Frequency domain signals are 
     * automatically computed. 
     * @param absorption is a floating point array that contains the absorption channel
     * @param dispersion is a floating point array that contains the dispersion channel
     */
    public void setSpectralDataPointsTD( float[] absorption, float[] dispersion )
    {
        m_absorptionChannelTD = new float[absorption.length];
        m_dispersionChannelTD = new float[dispersion.length];
        
        m_numberOfSamplePoints = absorption.length;
        
        for ( int i = 0; i < absorption.length ; i++ )
        {
            m_absorptionChannelTD[ i ] = absorption[ i ];
            m_dispersionChannelTD[ i ] = dispersion[ i ];
        }
                
        calculateFDRepresentation( );
        
         // Set a default xaxis (time domain)
        autoSetXaxis( );
    }

    /**
     * Set spectral data points.
     * @param absorption is a floating point array that contains the absorption channel
     * @param dispersion is a floating point array that contains the dispersion channel
     */
    public void setSpectralDataPointsTD( double[] absorption, double[] dispersion )
    {
        m_absorptionChannelTD = new float[absorption.length];
        m_dispersionChannelTD = new float[dispersion.length];
        
        m_numberOfSamplePoints = absorption.length;
        
        for ( int i = 0; i < absorption.length ; i++ )
        {
            m_absorptionChannelTD[ i ] = (float)absorption[ i ];
            m_dispersionChannelTD[ i ] = (float)dispersion[ i ];
        }

        calculateFDRepresentation( );

         // Set a default xaxis (time domain)
        autoSetXaxis( );
    }


    /** 
     * Set spectral data points.
     * @param absorption is a floating point array that contains the absorption channel
     * @param dispersion is a floating point array that contains the dispersion chammel
     */
    public void setSpectralDataPointsFD( float[] absorption, float[] dispersion )
    {
        m_absorptionChannelFD = new float[absorption.length];
        m_dispersionChannelFD = new float[dispersion.length];
        
        m_numberOfSamplePoints = absorption.length;
        
        for ( int i = 0; i < absorption.length ; i++ )
        {
            m_absorptionChannelFD[ i ] = absorption[ i ];
            m_dispersionChannelFD[ i ] = dispersion[ i ];
        }
        
        calculateTDRepresentation( );
        
         // Set a default xaxis (time domain)
        autoSetXaxis( );
    }

    /**
     * Set spectral data points.
     * @param absorption is a floating point array that contains the absorption channel
     * @param dispersion is a floating point array that contains the dispersion chammel
     */
    public void setSpectralDataPointsFD( double[] absorption, double[] dispersion )
    {
        m_absorptionChannelFD = new float[absorption.length];
        m_dispersionChannelFD = new float[dispersion.length];

        m_numberOfSamplePoints = absorption.length;
        
        for ( int i = 0; i < absorption.length ; i++ )
        {
            m_absorptionChannelFD[ i ] = (float)absorption[ i ];
            m_dispersionChannelFD[ i ] = (float)dispersion[ i ];
        }

        calculateTDRepresentation( );

         // Set a default xaxis (time domain)
        autoSetXaxis( );
    }

    /** 
     * Returns the time domain absorption channel.
     * @return returns the absorption channel floating point array
     */
    public float[] getAbsorptionChannelTD( )
    {
        if ( m_absorptionChannelTD != null )
        {
            this.autoSetXaxis( );
        }
        return m_absorptionChannelTD;
    }
    
    /** 
     * Returns the frequency domain absorption channel.
     * @return returns the absorption channel floating point array
     */
    public float[] getAbsorptionChannelFD( )
    {
        this.autoSetXaxis( );
        return m_absorptionChannelFD;
    }
    
    
    /** 
     * Returns the time domain absorption channel as Curve object.
     * @return the time domain absorption channel as Curve object
     */
    public Curve getAbsorptionCurveTDPoints( )
    {   
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, m_absorptionChannelTD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the time domain absorption channel as Curve object.
     * @return the time domain absorption channel as Curve object
     */
    public Curve getAbsorptionCurveTDSeconds( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, m_absorptionChannelTD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    
    /** 
     * Returns the frequency domain absorption channel as Curve object.
     * @return the frequency domain absorption channel as Curve object
     */
    public Curve getAbsorptionCurveFDPoints( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, m_absorptionChannelFD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    /** 
     * Returns the frequency domain absorption channel as Curve object.
     * @return the frequency domain absorption channel as Curve object
     */
    public Curve getAbsorptionCurveFDHertz( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, m_absorptionChannelFD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    
    /** 
     * Returns the frequency domain absorption channel as Curve object.
     * @return the frequency domain absorption channel as Curve object
     */
    public Curve getAbsorptionCurveFDPPM( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, m_absorptionChannelFD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    /** 
     * Returns the time domain dispersion channel.
     * @return returns the dispersion channel floating point array
     */
    public float[] getDispersionChannelTD( )
    {
        this.autoSetXaxis( );
        return m_dispersionChannelTD;
    }    
    
    /** 
     * Returns the frequency domain dispersion channel.
     * @return returns the dispersion channel floating point array
     */
    public float[] getDispersionChannelFD( )
    {
        this.autoSetXaxis( );
        return m_dispersionChannelFD;
    }    
    
    /** 
     * Returns the time domain dispersion channel as Curve object.
     * @return the time domain dispersion channel as Curve object
     */
    public Curve getDispersionCurveTDPoints( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, m_dispersionChannelTD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the time domain dispersion channel as Curve object.
     * @return the time domain dispersion channel as Curve object
     */
    public Curve getDispersionCurveTDSeconds( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, m_dispersionChannelTD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }    
    
    /** 
     * Returns the frequency domain dispersion channel as Curve object.
     * @return the frequency domain dispersion channel as Curve object
     */
    public Curve getDispersionCurveFDPoints( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, m_dispersionChannelFD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    } 

    /** 
     * Returns the frequency domain dispersion channel as Curve object.
     * @return the frequency domain dispersion channel as Curve object
     */
    public Curve getDispersionCurveFDHertz( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, m_dispersionChannelFD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    } 
    
    /** 
     * Returns the frequency domain dispersion channel as Curve object.
     * @return the frequency domain dispersion channel as Curve object
     */
    public Curve getDispersionCurveFDPPM( )
    {
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, m_dispersionChannelFD, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }    
    
    /** 
     * Returns the time domain magnitude of the (absorption,dispersion) channel.
     * @return the time domain magnitude floating point array
     */
    public Curve getMagnitudeCurveTDPoints( )
    {
        float[] magnitude = new float[ m_absorptionChannelTD.length ];
        
        for ( int i= 0; i < m_absorptionChannelTD.length; i++ )
        {
            magnitude[i] = (float)Math.sqrt( m_absorptionChannelTD[i] * m_absorptionChannelTD[i] + 
                                             m_dispersionChannelTD[i] * m_dispersionChannelTD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, magnitude, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the time domain magnitude of the (absorption,dispersion) channel.
     * @return the time domain magnitude floating point array
     */
    public Curve getMagnitudeCurveTDSeconds( )
    {
        float[] magnitude = new float[ m_absorptionChannelTD.length ];
        
        for ( int i= 0; i < m_absorptionChannelTD.length; i++ )
        {
            magnitude[i] = (float)Math.sqrt( m_absorptionChannelTD[i] * m_absorptionChannelTD[i] + 
                                             m_dispersionChannelTD[i] * m_dispersionChannelTD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, magnitude, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    /** 
     * Returns the frequency domain magnitude of the 
     * (absorption,dispersion) channel.
     * @return returns the frequency domain magnitude floating point array
     */
    public Curve getMagnitudeCurveFDPoints( )
    {
        float[] magnitude = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            magnitude[i] = (float)Math.sqrt( m_absorptionChannelFD[i] * m_absorptionChannelFD[i] + 
                                             m_dispersionChannelFD[i] * m_dispersionChannelFD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, magnitude, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    /** 
     * Returns the frequency domain magnitude of the 
     * (absorption,dispersion) channel.
     * @return returns the frequency domain magnitude floating point array
     */
    public Curve getMagnitudeCurveFDHertz( )
    {
        float[] magnitude = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            magnitude[i] = (float)Math.sqrt( m_absorptionChannelFD[i] * m_absorptionChannelFD[i] + 
                                             m_dispersionChannelFD[i] * m_dispersionChannelFD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, magnitude, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the frequency domain magnitude of the 
     * (absorption,dispersion) channel.
     * @return returns the frequency domain magnitude floating point array
     */
    public Curve getMagnitudeCurveFDPPM( )
    {
        float[] magnitude = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            magnitude[i] = (float)Math.sqrt( m_absorptionChannelFD[i] * m_absorptionChannelFD[i] + 
                                             m_dispersionChannelFD[i] * m_dispersionChannelFD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, magnitude, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    /** 
     * Returns the power of the (absorption,dispersion) channel.
     * @return returns the power floating point array
     */
    public Curve getPowerCurveTDPoints( )
    {
        float[] power = new float[ m_absorptionChannelTD.length ];
        
        for ( int i= 0; i < m_absorptionChannelTD.length; i++ )
        {   
            power[i] =  m_absorptionChannelTD[i] * m_absorptionChannelTD[i] + 
                        m_dispersionChannelTD[i] * m_dispersionChannelTD[i];
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, power, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }    

    /** 
     * Returns the power of the (absorption,dispersion) channel.
     * @return returns the power floating point array
     */
    public Curve getPowerCurveTDSeconds( )
    {
        float[] power = new float[ m_absorptionChannelTD.length ];
        
        for ( int i= 0; i < m_absorptionChannelTD.length; i++ )
        {   
            power[i] =  m_absorptionChannelTD[i] * m_absorptionChannelTD[i] + 
                        m_dispersionChannelTD[i] * m_dispersionChannelTD[i];
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, power, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the frequency domain power of the 
     * (absorption,dispersion) channel.
     * @return the frequency domain power floating point array
     */
    public Curve getPowerCurveFDPoints( )
    {
        float[] power = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            power[i] =  m_absorptionChannelFD[i] * m_absorptionChannelFD[i] + 
                        m_dispersionChannelFD[i] * m_dispersionChannelFD[i];
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, power, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }    

    /** 
     * Returns the frequency domain power of the 
     * (absorption,dispersion) channel.
     * @return the frequency domain power floating point array
     */
    public Curve getPowerCurveFDHertz( )
    {
        float[] power = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            power[i] =  m_absorptionChannelFD[i] * m_absorptionChannelFD[i] + 
                        m_dispersionChannelFD[i] * m_dispersionChannelFD[i];
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, power, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }    
    
    /** 
     * Returns the frequency domain power of the 
     * (absorption,dispersion) channel.
     * @return the frequency domain power floating point array
     */
    public Curve getPowerCurveFDPPM( )
    {
        float[] power = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            power[i] =  m_absorptionChannelFD[i] * m_absorptionChannelFD[i] + 
                        m_dispersionChannelFD[i] * m_dispersionChannelFD[i];
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, power, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }    
    
    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getPhaseCurveTDPoints( )
    {
        float[] phase = new float[ m_absorptionChannelTD.length ];
        
        for ( int i= 0; i < m_absorptionChannelTD.length; i++ )
        {
            phase[i] = (float)Math.atan2( m_absorptionChannelTD[i] , m_dispersionChannelTD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, phase, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getPhaseCurveTDSeconds( )
    {
        float[] phase = new float[ m_absorptionChannelTD.length ];
        
        for ( int i= 0; i < m_absorptionChannelTD.length; i++ )
        {
            phase[i] = (float)Math.atan2( m_absorptionChannelTD[i] , m_dispersionChannelTD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisTD, phase, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }

    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getPhaseCurveFDPoints( )
    {
        float[] phase = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            phase[i] = (float)Math.atan2( m_absorptionChannelFD[i] , m_dispersionChannelFD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, phase, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    
    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getPhaseCurveFDHertz( )
    {
        float[] phase = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            phase[i] = (float)Math.atan2( m_absorptionChannelFD[i] , m_dispersionChannelFD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, phase, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getPhaseCurveFDPPM( )
    {
        float[] phase = new float[ m_absorptionChannelFD.length ];
        
        for ( int i= 0; i < m_absorptionChannelFD.length; i++ )
        {
            phase[i] = (float)Math.atan2( m_absorptionChannelFD[i] , m_dispersionChannelFD[i] );
        }
        this.autoSetXaxis( );
        return new Curve( m_xaxisFD, phase, m_curveColor, m_curveDrawingStyle, m_curveLabel );
    }
    
    
    /** 
     * Returns the complex signal with alternating absorption point
     * dispersion point.
     * @return returns float array containing the sampled spectroscopic data points
     */
    public float[] getBothChannelsTD( )
    {
        return m_bothChannelsTD;
    }
    
    /** 
     * Sets the sample frequency attribute.
     * @param sampleFrequency Frequency with which the MR signal is acquired 
     */    
    public void setSampleFrequency( float sampleFrequency )
    {
        m_acqFrequency = sampleFrequency;
    }
    
    /** 
     * Returns the sample frequency with which the MR signal was acquired.
     * @return Returns the sample frequency in Hz 
     */    
    public float getSampleFrequency( )
    {
        return m_acqFrequency;
    }
    
    
    /** 
     * Sets the sample frequency attribute.
     * @param samplePoints 
     */    
    public void setNumberOfSamplePoints( int samplePoints )
    {
        m_numberOfSamplePoints = samplePoints;
    }
    
    /** 
     * Returns the sample frequency with which the MR signal was acquired.
     * @return Returns the sample frequency in Hz 
     */    
    public int getNumberOfSamplePoints( )
    {
        return m_numberOfSamplePoints;
    }
    
    /**
     * Clears both time and frequency domain channels. 
     */
    public void clear( )
    {
        for ( int i=0; i < m_absorptionChannelTD.length; i++ )
        {
            m_absorptionChannelTD[i] = 0f;
            m_dispersionChannelTD[i] = 0f;
            m_absorptionChannelFD[i] = 0f;
            m_dispersionChannelFD[i] = 0f;
        }
        
        for ( int i=0; i < 2*m_absorptionChannelTD.length; i++ )
        {
            m_bothChannelsTD[i] = 0f;
            m_bothChannelsFD[i] = 0f;
        }
    }

    /**
     * Writes attributes to the open RandomAccessFile.
     * @param out
     */
    public void write( RandomAccessFile out )
    {
        try
        {
            out.writeUTF( "<SingleVoxelSpectralData>" );
            out.writeUTF( "<SingleVoxelSpectralDataHeader>" );
            out.writeInt( m_absorptionChannelTD.length );
            out.writeFloat( m_acqFrequency );
            out.writeUTF( "</SingleVoxelSpectralDataHeader>" );
            
            out.writeUTF( "<AbsorptionChannelTD>" );
            if ( m_absorptionChannelTD != null )
            {
                m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 2.0f;
            
                for ( int i = 0; i < this.m_absorptionChannelTD.length; i++ )
                {
                    out.writeFloat( m_absorptionChannelTD[i] );
                }
                
                 m_absorptionChannelTD[0] = m_absorptionChannelTD[0] / 2.0f;
            }
            out.writeUTF( "</AbsorptionChannelTD>" );
            out.writeUTF( "<DispersionChannelTD>" );
            if ( m_dispersionChannelTD != null )
            {
                
                m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 2.0f;
                
                for ( int i = 0; i < this.m_absorptionChannelTD.length;i++ )
                {
                    out.writeFloat( m_dispersionChannelTD[i] );
                }
                
                m_dispersionChannelTD[0] = m_dispersionChannelTD[0] / 2.0f;
            }
            out.writeUTF( "</DispersionChannelTD>" );
            out.writeUTF( "</SingleVoxelSpectralData>" );
        }
        catch ( IOException e )
        {
            System.out.println( "Error writing spectral response to file." );
        }
    }
    
    /**
     * Reads and parses the attributes from a RandomAccesFile that contains
     * the response part of the object.
     * @param in
     * @param skipFirstTag
     */
    public void read( RandomAccessFile in, boolean skipFirstTag )
    {
        try
        {
            String tag = "<SingleVoxelSpectralData>";
            
            if ( !skipFirstTag == true )
            {    
                tag = in.readUTF( );
            }
            // System.out.println( "tag = " + tag );
  
            if ( tag.compareTo( "<SingleVoxelSpectralData>" ) == 0 )
            {
                tag = in.readUTF( );
                // System.out.println( "tag = " + tag );
                if ( tag.compareTo( "<SingleVoxelSpectralDataHeader>" ) == 0 )
                {
                    
                    m_numberOfSamplePoints = in.readInt( );
                    
                    // Allocate memory for the data arrays
                    allocateArrays( m_numberOfSamplePoints );
                    m_acqFrequency = in.readFloat( );
                    
                }
                tag = in.readUTF( );
                // System.out.println( "tag = " + tag );
            }
            
            tag = in.readUTF( );
            // System.out.println( "tag = " + tag );
            if ( tag.compareTo( "<AbsorptionChannelTD>" ) == 0 )
            {
                
                m_absorptionChannelTD = new float[ m_numberOfSamplePoints ];
                for ( int i = 0 ; i < m_numberOfSamplePoints; i++ )
                {
                    m_absorptionChannelTD[i] = in.readFloat( );
                }
                // read closing tag absorption channel
                tag = in.readUTF( );
                // System.out.println( "tag = " + tag );
            }
            
            tag = in.readUTF( );
            // System.out.println( "tag = " + tag );
            if ( tag.compareTo( "<DispersionChannelTD>" ) == 0 )
            {
                
                m_dispersionChannelTD = new float[ m_numberOfSamplePoints ];
                for ( int i = 0 ; i < m_numberOfSamplePoints; i++ )
                {
                    m_dispersionChannelTD[i] = in.readFloat( );
                }
                // read closing tag dispersion channel
                tag = in.readUTF( );
                // System.out.println( "tag = " + tag );
            }
            
            m_absorptionChannelTD[0] = m_absorptionChannelTD[0]/2.0f;
            m_dispersionChannelTD[0] = m_dispersionChannelTD[0]/2.0f;
            
            // Read the closing single voxel spectral data tag
            tag = in.readUTF( );
            //   + tag );
        }
        catch ( IOException e )
        {
            e.printStackTrace( );
        }
        
        // brings the frequency domain response into sync
        calculateFDRepresentation( );
    }
    
    /**
     * Writes the current file to TDFDFit file format.
     * @param fileName
     */
    public void writeAsTdfdfitFile( String fileName )
    {
        
        TDFDFittingDefaults fitDef = TDFDFittingDefaults.getInstance( );
        fitDef.load( );
        
        String path = fitDef.getProperty( "Fit project path" );
        fileName = path + System.getProperty( "file.separator" ) + fileName;
        System.out.println( "fileName : " + fileName );
        try
        {
            
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            
            for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                buf.putFloat( m_absorptionChannelFD[i] );
                
                writer.write( buf.get( 3 ) );
                writer.write( buf.get( 2 ) );
                writer.write( buf.get( 1 ) );
                writer.write( buf.get( 0 ) );
                
                buf1 = ByteBuffer.allocate( 4 );
                
                buf1.putFloat( m_dispersionChannelFD[i] );                
                
                writer.write( buf1.get( 3 ) );
                writer.write( buf1.get( 2 ) );
                writer.write( buf1.get( 1 ) );
                writer.write( buf1.get( 0 ) );
                                
            }

            writer.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted measured datafile" );
        }
    }

    /**
     * Writes the current file to TDFDFit file format.
     * @param path
     * @param fileName
     */
    public void writeAsTdfdfitFile( String path, String fileName )
    {
        
        if ( path != null )
        {
            fileName = path + System.getProperty( "file.separator" ) + fileName;
        }
        
        System.out.println( "fileName : " + fileName );
        try
        {
            
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            
            for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                buf.putFloat( m_absorptionChannelFD[i] );
                
                writer.write( buf.get( 3 ) );
                writer.write( buf.get( 2 ) );
                writer.write( buf.get( 1 ) );
                writer.write( buf.get( 0 ) );
                
                buf1 = ByteBuffer.allocate( 4 );
                
                buf1.putFloat( m_dispersionChannelFD[i] );                
                
                writer.write( buf1.get( 3 ) );
                writer.write( buf1.get( 2 ) );
                writer.write( buf1.get( 1 ) );
                writer.write( buf1.get( 0 ) );
                                
            }

            writer.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted measured datafile" );
        }
    }
    
    /**
     * Reads a TDFDFit file format binary file and stores its contents in the
     * appropriate member variables.
     * @param numberOfComplexPoints
     * @param fileRef
     */
    public void readTDFDFitFile( int numberOfComplexPoints, String fileRef )
    {
        
        try
        {          
            RandomAccessFile reader = new RandomAccessFile( fileRef, "rw" );
            
            ByteBuffer buf;
            byte byte0;
            byte byte1;
            byte byte2;
            byte byte3;
            float readFloat;
            
            for ( int i = 0; i < numberOfComplexPoints; i++ )
            {
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = reader.readByte( );
                byte1 = reader.readByte( );
                byte2 = reader.readByte( );
                byte3 = reader.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_absorptionChannelFD[i] = readFloat;
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = reader.readByte( );
                byte1 = reader.readByte( );
                byte2 = reader.readByte( );
                byte3 = reader.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_dispersionChannelFD[i] = readFloat;
                
            }
            // the frequency domain response is read so generate the time 
            // domain response to get full representation
            calculateTDRepresentation( );
            
            reader.close( );
            
        }
        catch ( IOException e )
        {
            e.printStackTrace( );
            System.out.println( "3. Cannot read tdfdfit-formatted measured datafile" );
        }
        
        autoSetXaxis( );
    }
    
    /**
     * Sets the color of the curve to be drawn.
     * @param color
     */
    public void setCurveColor( Color color )
    {
        m_curveColor = color;
    }

    /**
     * Sets the color of the curve to be drawn.
     * @param label
     */
    public void setCurveLabel( String label )
    {
        m_curveLabel = label;
    }

    /**
     * Sets the color of the curve to be drawn.
     * @param style
     */
    public void setCurveDrawingStyle( int style )
    {
        this.m_curveDrawingStyle = style;
    }

    /**
     * Sets the DICOM file header information.
     * @param fileInfo
     */
    public void setDicomHeaderInfo( FileInfo fileInfo )
    {
        m_dicomHeaderInfo = fileInfo;
    }
    
    /**
     * Returns the DICOM file header information.
     * @return 
     */
    public FileInfo getDicomHeaderInfo( )
    {
        return m_dicomHeaderInfo;
    }
    
    
    /**
     * Returns the resonance frequency of the scanner/spectrometer.
     * @return
     */
    public long getResonanceFrequency( )
    {
        return m_resonanceFrequency;
    }

    /**
     * Returns the recorded nucleus type.
     * @return
     */
    public String getNucleusType( )
    {
        return m_nucluesType;
    }
    
    /**
     * Sets the recorded nucleus type.
     * @param nucleusType
     */
    public void setNucleusType( String nucleusType )
    {
        m_nucluesType = nucleusType;
    }
    
    /**
     * Sets the LinkedHashMap containing all Siemens specific header 
     * information.
     * @param ssd
     */
    public void setSiemensSpecificDicomHeaderInfo( LinkedHashMap ssd )
    {
        
        m_siemensHeaderInfo = ssd;
        
        String name;
        
        if ( ssd != null )
        {
            Iterator it = ssd.keySet( ).iterator( );
        
            while ( it.hasNext( ) )
            {
                Object key = it.next( );
                
                // System.out.println( "" + (String)key );
                if ( ((String)key).compareTo(  "sTXSPEC.asNucleusInfo[0].lFrequency" ) == 0 )
                {
                    m_resonanceFrequency = new Long( (String)ssd.get( key ) ).longValue( );
                }
                // System.out.println( "" + (String)key );
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.dThickness" ) == 0 )
                {
                    m_sVOIThickness = new Double( (String)ssd.get( key ) );
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.dReadoutFOV" ) == 0 )
                {
                    m_sVoIPhaseFOV = new Double( (String)ssd.get( key ) );
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.dPhaseFOV" ) == 0 )
                {
                    m_sVoIReadoutFOV = new Double( (String)ssd.get( key ) );
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sPosition.dSag" ) == 0 )
                {
                    if ( ((String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dSagVoxelPosition =  d.doubleValue( );
                    }
                    else
                    {
                        m_dSagVoxelPosition = 0.00;
                    }
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sPosition.dCor" ) == 0 )
                {
                    if ( ((String)ssd.get( key )).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dCorVoxelPosition = d.doubleValue( );
                    }
                    else
                    {
                        m_dCorVoxelPosition = 0.00;
                    }
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sPosition.dTra" ) == 0 )
                {
                    if ( ( (String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dTraVoxelPosition = d.doubleValue( );
                    }
                    else
                    {
                        m_dTraVoxelPosition = 0.00;
                    }
                }

                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sNormal.dSag" ) == 0 )
                { 
                    if ( ((String)ssd.get( key )).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dSagVoxelOrientation = d.doubleValue( );
                    }
                    else
                    {
                        m_dSagVoxelOrientation = 0.00;
                    }
                
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sNormal.dCor" ) == 0 )
                {
                    if ( ((String)ssd.get( key )).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dCorVoxelOrientation = d.doubleValue( );
                    }
                    else
                    {
                        m_dCorVoxelOrientation = 0.00;
                    }
                    
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sNormal.dTra" ) == 0 )
                {
                    if ( ( (String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dTraVoxelOrientation = d.doubleValue( );
                    }
                    else
                    {
                        m_dTraVoxelOrientation = 0.00;
                    }
                }
                
                if ( ( (String)key).compareTo( "sSpecPara.sVoI.dInPlaneRot" ) == 0 )
                {
                    if ( ( (String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dInPlaneRotation = d.doubleValue( );
                    }
                    else
                    {
                        m_dInPlaneRotation = 0.0;
                    }
                }
               
                if ( ( (String)key).compareTo( "alTE[0]" ) == 0 )
                {
                    m_TE = ( new Long( (String)ssd.get( key ) ) ).intValue( );
                }
                        
                if ( ( (String)key).compareTo( "sTXSPEC.asNucleusInfo[0].lFrequency" ) == 0 )
                {
                    m_resonanceFrequency = ( new Long( (String)ssd.get( key ) ) ).longValue( );
                }

                if ( ( (String)key).compareTo( "sRXSPEC.alDwellTime[0]" ) == 0 )
                {
                    m_sampleDwellTime = ( new Integer( (String)ssd.get( key ) ) ).intValue( );
                    
                    m_acqFrequency =  1000f * 1000000f / ( 2f * m_sampleDwellTime );
                    
                }
                
                if ( ( (String)key).compareTo( "sTXSPEC.asNucleusInfo[0].tNucleus" ) == 0 )
                {
                    m_nucluesType = (String)ssd.get( key );
                }
            } 
            
            autoSetXaxis( ); 
            
        }
        else
        {
            // assume 1kHz bandwidth just to have a x scale
            MRSignalProperties m_sigProp = MRSignalProperties.getInstance( );
            m_acqFrequency = m_sigProp.getFrequencyScale( ) * m_sigProp.getNumberOfSamplePoints( );
            autoSetXaxis( );
        }
        
        
        if ( ( this.m_dTraVoxelOrientation > this.m_dCorVoxelOrientation ) &&
             ( this.m_dTraVoxelOrientation > this.m_dSagVoxelOrientation ) )
        {
            computeVoxelForTransversalPrescription( );
        }
        
        if ( ( this.m_dCorVoxelOrientation > this.m_dTraVoxelOrientation ) &&
             ( this.m_dCorVoxelOrientation > this.m_dSagVoxelOrientation ) )
        {
            computeVoxelForCoronalPrescription( );
        }
        
        if ( ( this.m_dSagVoxelOrientation > this.m_dTraVoxelOrientation ) &&
             ( this.m_dSagVoxelOrientation > this.m_dCorVoxelOrientation ) )
        {
            computeVoxelForSagitalPrescription( );
        }
        
    }

    
    /**
     * Computes a voxel with sagital prescription.
     */
    private void computeVoxelForSagitalPrescription( )
    {

        Point3D rotAxis1 = new Point3D( m_dCorVoxelOrientation, -m_dTraVoxelOrientation, 0.0 );
        rotAxis1 = rotAxis1.normalize( rotAxis1 );

        
        Point3D rotAxis2 = new Point3D( m_dSagVoxelOrientation, m_dCorVoxelOrientation, m_dTraVoxelOrientation );
        rotAxis2 = rotAxis2.normalize( rotAxis2 );

        
        double rotAngle1 = -Math.acos( m_dSagVoxelOrientation );
        double rotAngle2 = -m_dInPlaneRotation;
        
        // double rotAngle1 = 0.0;
        // double rotAngle2 = 0.0;
        // should be Point3D translate = new Point3D( m_dCorVoxelPosition, m_dSagVoxelPosition, m_dTraVoxelPosition );
        
        Point3D translate = new Point3D( m_dSagVoxelPosition, m_dCorVoxelPosition, m_dTraVoxelPosition );
        
        
        Point3D point1 = new Point3D(  - m_sVOIThickness / 2.,
                                         m_sVoIReadoutFOV / 2. ,
                                       - m_sVoIPhaseFOV / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point1 = point1.rotate3D( point1, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point1 = point1.rotate3D( point1, rotAxis2, rotAngle2 );
            }
            else
            {
                ;
            }
        }
        point1 = point1.translate( translate );
        
        
        Point3D point2 = new Point3D(    m_sVOIThickness / 2.,
                                         m_sVoIReadoutFOV / 2. ,
                                       - m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point2 = point2.rotate3D( point2, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point2 = point2.rotate3D( point2, rotAxis2, rotAngle2 );
            }
        }
        point2 = point2.translate( translate );
        

        Point3D point3 = new Point3D( m_sVOIThickness / 2.,
                                      m_sVoIReadoutFOV / 2. ,
                                      m_sVoIPhaseFOV / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point3 = point3.rotate3D( point3, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point3 = point3.rotate3D( point3, rotAxis2, rotAngle2 );
            }
        }
      
        point3 = point3.translate( translate );
        
        
        Point3D point4 = new Point3D( - m_sVOIThickness / 2.,
                                        m_sVoIReadoutFOV / 2. ,
                                        m_sVoIPhaseFOV / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point4 = point4.rotate3D( point4, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point4 = point4.rotate3D( point4, rotAxis2, rotAngle2 );
            }
        }
        
        point4 = point4.translate( translate );

        Point3D point5 = new Point3D( - m_sVOIThickness / 2.,
                                      - m_sVoIReadoutFOV / 2. ,
                                      - m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point5 = point5.rotate3D( point5, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point5 = point5.rotate3D( point5, rotAxis2, rotAngle2 );
            }
        }
        
        point5 = point5.translate( translate );
        
        
        Point3D point6 = new Point3D(   m_sVOIThickness / 2.,
                                      - m_sVoIReadoutFOV / 2. ,
                                      - m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point6 = point6.rotate3D( point6, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point6 = point6.rotate3D( point6, rotAxis2, rotAngle2 );
            }
        }
        
        point6 = point6.translate( translate );
        

        Point3D point7 = new Point3D(   m_sVOIThickness / 2.,
                                      - m_sVoIReadoutFOV / 2. ,
                                        m_sVoIPhaseFOV / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point7 = point7.rotate3D( point7, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point7 = point7.rotate3D( point7, rotAxis2, rotAngle2 );
            }
        }
        
        point7 = point7.translate( translate );
        
        Point3D point8 = new Point3D( - m_sVOIThickness / 2.,
                                      - m_sVoIReadoutFOV / 2. ,
                                        m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point8 = point8.rotate3D( point8, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point8 = point8.rotate3D( point8, rotAxis2, rotAngle2 );
            }
        }
        
        point8 = point8.translate( translate );
        
        m_voxelGeometry = new Spectroscopic3DVoxel( point1, point2, point3, point4,
                                                    point5, point6, point7, point8 );
    }
    
    /**
     * Computes a voxel with Transversal prescription.
     */
    private void computeVoxelForCoronalPrescription( )
    {

        Point3D rotAxis1 = new Point3D( m_dTraVoxelOrientation, -m_dSagVoxelOrientation, 0.0 );
        rotAxis1 = rotAxis1.normalize( rotAxis1 );

        
        Point3D rotAxis2 = new Point3D( m_dSagVoxelOrientation, m_dCorVoxelOrientation, m_dTraVoxelOrientation );
        rotAxis2 = rotAxis2.normalize( rotAxis2 );

        
        double rotAngle1 = -Math.acos( m_dCorVoxelOrientation );
        double rotAngle2 = -m_dInPlaneRotation;
        
        // double rotAngle1 = 0.0;
        // double rotAngle2 = 0.0;
        // should be Point3D translate = new Point3D( m_dCorVoxelPosition, m_dSagVoxelPosition, m_dTraVoxelPosition );
        
        Point3D translate = new Point3D( m_dSagVoxelPosition, m_dCorVoxelPosition, m_dTraVoxelPosition );
        
        
        Point3D point1 = new Point3D(  - m_sVoIReadoutFOV / 2., 
                                         m_sVOIThickness / 2.,
                                       - m_sVoIPhaseFOV / 2. );
        
        
        if ( rotAngle1 != 0.0 )
        {
            point1 = point1.rotate3D( point1, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point1 = point1.rotate3D( point1, rotAxis2, rotAngle2 );
            }
            else
            {
                ;
            }
        }
        point1 = point1.translate( translate );
        
        Point3D point2 = new Point3D(    m_sVoIReadoutFOV / 2., 
                                         m_sVOIThickness / 2.,
                                       - m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point2 = point2.rotate3D( point2, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point2 = point2.rotate3D( point2, rotAxis2, rotAngle2 );
            }
        }
        point2 = point2.translate( translate );
        

        Point3D point3 = new Point3D(    m_sVoIReadoutFOV / 2., 
                                         m_sVOIThickness / 2.,
                                         m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point3 = point3.rotate3D( point3, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point3 = point3.rotate3D( point3, rotAxis2, rotAngle2 );
            }
        }
      
        point3 = point3.translate( translate );
        
        Point3D point4 = new Point3D(  - m_sVoIReadoutFOV / 2., 
                                         m_sVOIThickness / 2.,
                                         m_sVoIPhaseFOV / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point4 = point4.rotate3D( point4, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point4 = point4.rotate3D( point4, rotAxis2, rotAngle2 );
            }
        }
        
        point4 = point4.translate( translate );
        

        Point3D point5 = new Point3D(  - m_sVoIReadoutFOV / 2., 
                                       - m_sVOIThickness / 2.,
                                       - m_sVoIPhaseFOV / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point5 = point5.rotate3D( point5, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point5 = point5.rotate3D( point5, rotAxis2, rotAngle2 );
            }
        }
        
        point5 = point5.translate( translate );
        
        Point3D point6 = new Point3D(    m_sVoIReadoutFOV / 2., 
                                      -  m_sVOIThickness / 2.,
                                      -  m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point6 = point6.rotate3D( point6, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point6 = point6.rotate3D( point6, rotAxis2, rotAngle2 );
            }
        }
        
        point6 = point6.translate( translate );
        
        Point3D point7 = new Point3D(    m_sVoIReadoutFOV / 2., 
                                      -  m_sVOIThickness / 2.,
                                         m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point7 = point7.rotate3D( point7, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point7 = point7.rotate3D( point7, rotAxis2, rotAngle2 );
            }
        }
        
        point7 = point7.translate( translate );
        
        Point3D point8 = new Point3D( -  m_sVoIReadoutFOV / 2., 
                                      -  m_sVOIThickness / 2.,
                                         m_sVoIPhaseFOV / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point8 = point8.rotate3D( point8, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point8 = point8.rotate3D( point8, rotAxis2, rotAngle2 );
            }
        }
        
        point8 = point8.translate( translate );
        
        m_voxelGeometry = new Spectroscopic3DVoxel( point1, point2, point3, point4,
                                                    point5, point6, point7, point8 );
    }
 
    /**
     * Computes a voxel with Transversal prescription.
     */
    private void computeVoxelForTransversalPrescription( )
    {

        Point3D rotAxis1 = new Point3D( m_dCorVoxelOrientation, -m_dSagVoxelOrientation, 0.0 );
        rotAxis1 = rotAxis1.normalize( rotAxis1 );
        
        Point3D rotAxis2 = new Point3D( m_dSagVoxelOrientation, m_dCorVoxelOrientation, m_dTraVoxelOrientation );
        rotAxis2 = rotAxis2.normalize( rotAxis2 );

        
        double rotAngle1 = -Math.acos( m_dTraVoxelOrientation );
        double rotAngle2 = -m_dInPlaneRotation;
        
        // double rotAngle1 = 0.0;
        // double rotAngle2 = 0.0;
        // should be Point3D translate = new Point3D( m_dCorVoxelPosition, m_dSagVoxelPosition, m_dTraVoxelPosition );
        
        Point3D translate = new Point3D( m_dSagVoxelPosition, m_dCorVoxelPosition, m_dTraVoxelPosition );
        
        Point3D point1 = new Point3D(  - m_sVoIPhaseFOV / 2., 
                                         m_sVoIReadoutFOV / 2.,
                                       - m_sVOIThickness / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point1 = point1.rotate3D( point1, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point1 = point1.rotate3D( point1, rotAxis2, rotAngle2 );
            }
            else
            {
                ;
            }
        }
        point1 = point1.translate( translate );
        
        
        Point3D point2 = new Point3D(   m_sVoIPhaseFOV / 2., 
                                        m_sVoIReadoutFOV / 2.,
                                      - m_sVOIThickness / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point2 = point2.rotate3D( point2, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point2 = point2.rotate3D( point2, rotAxis2, rotAngle2 );
            }
        }
        point2 = point2.translate( translate );
        
        
        Point3D point3 = new Point3D(   m_sVoIPhaseFOV / 2., 
                                        m_sVoIReadoutFOV / 2.,
                                        m_sVOIThickness / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point3 = point3.rotate3D( point3, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point3 = point3.rotate3D( point3, rotAxis2, rotAngle2 );
            }
        }
      
        point3 = point3.translate( translate );
        
        
        Point3D point4 = new Point3D(  - m_sVoIPhaseFOV / 2., 
                                         m_sVoIReadoutFOV / 2.,
                                         m_sVOIThickness / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point4 = point4.rotate3D( point4, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point4 = point4.rotate3D( point4, rotAxis2, rotAngle2 );
            }
        }
        
        point4 = point4.translate( translate );
        
        Point3D point5 = new Point3D(  - m_sVoIPhaseFOV / 2., 
                                       - m_sVoIReadoutFOV / 2.,
                                       - m_sVOIThickness / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point5 = point5.rotate3D( point5, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point5 = point5.rotate3D( point5, rotAxis2, rotAngle2 );
            }
        }
        
        point5 = point5.translate( translate );
        
        Point3D point6 = new Point3D(   m_sVoIPhaseFOV / 2., 
                                      - m_sVoIReadoutFOV / 2.,
                                      - m_sVOIThickness / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point6 = point6.rotate3D( point6, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point6 = point6.rotate3D( point6, rotAxis2, rotAngle2 );
            }
        }
        
        point6 = point6.translate( translate );
        
        Point3D point7 = new Point3D(   m_sVoIPhaseFOV / 2., 
                                      - m_sVoIReadoutFOV / 2.,
                                        m_sVOIThickness / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point7 = point7.rotate3D( point7, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point7 = point7.rotate3D( point7, rotAxis2, rotAngle2 );
            }
        }
        
        point7 = point7.translate( translate );
        
        Point3D point8 = new Point3D(  - m_sVoIPhaseFOV / 2., 
                                       - m_sVoIReadoutFOV / 2.,
                                         m_sVOIThickness / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point8 = point8.rotate3D( point8, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point8 = point8.rotate3D( point8, rotAxis2, rotAngle2 );
            }
        }
        
        point8 = point8.translate( translate );
        
        m_voxelGeometry = new Spectroscopic3DVoxel( point1, point2, point3, point4,
                                                    point5, point6, point7, point8 );
    }
 
    
    
    /**
     * Get single voxel geometry data.
     * @return 
     */
    public Spectroscopic3DVoxel getSpectroscopic3DVoxel( )
    {
        return m_voxelGeometry;
    }
    
    /**
     * Returns the LinkedHashMap containing all Siemens 
     * specific header information.
     * @return 
     */
    public LinkedHashMap getSiemensSpecificDicomHeaderInfo( )
    {
        return m_siemensHeaderInfo;
    }
    
    /**
     * Returns the string containing the voxel dimensions.
     * @return 
     */
    public String getVoxelDimensionsAsString( )
    {
        String voxelDimensions;
        
        NumberFormat format = new DecimalFormat( "###.#" );
        
        if ( m_siemensHeaderInfo != null )
        {
            voxelDimensions = "[ " +  m_siemensHeaderInfo.get( "sSpecPara.sVoI.dThickness" ) + ", ";
            voxelDimensions += m_siemensHeaderInfo.get( "sSpecPara.sVoI.dPhaseFOV" ) + ", ";
            voxelDimensions += m_siemensHeaderInfo.get( "sSpecPara.sVoI.dReadoutFOV" ) + " ]";
        }
        else
        { 
            voxelDimensions = "[ " + format.format( m_sVOIThickness ) + ", ";
            voxelDimensions += format.format( m_sVoIPhaseFOV ) + ", ";
            voxelDimensions += format.format( m_sVoIReadoutFOV ) + " ]";
        }
        return voxelDimensions;
    }

    /**
     * Returns the MEDIA_STORAGE_SOP_INSTANCE_UID of reference file 0.
     * @return 
     */
    public String getReferenceFile0_UID( )
    {
        String refFile0 = "not_defined";
        
        if ( (String)m_siemensHeaderInfo.get( "tReferenceImage0" ) != null )
        {
            refFile0 = ( (String)m_siemensHeaderInfo.get( "tReferenceImage0" ) ).replace( '"', ' ' ).trim( );
        }
        
        return refFile0;
    }

    /**
     * Returns the MEDIA_STORAGE_SOP_INSTANCE_UID of reference file 1.
     * @return 
     */
    public String getReferenceFile1_UID( )
    {
        String refFile1 = "not_defined";
    
        if ( (String)m_siemensHeaderInfo.get( "tReferenceImage1" ) != null )
        {
            refFile1 = ( (String)m_siemensHeaderInfo.get( "tReferenceImage1" ) ).replace( '"', ' ' ).trim( );
        }
        return refFile1;
    }

    /**
     * Returns the MEDIA_STORAGE_SOP_INSTANCE_UID of reference file 2.
     * @return 
     */
    public String getReferenceFile2_UID( )
    {
        String refFile2 = "not_defined";
        if ( (String)m_siemensHeaderInfo.get( "tReferenceImage2" ) != null )
        {
            refFile2 = ( (String)m_siemensHeaderInfo.get( "tReferenceImage2" ) ).replace( '"', ' ' ).trim( );
        }
        return refFile2;
    }
    
    /**
     * Sets the file reference of image reference dicom file 0.
     * (i.e. a file system path)
     * @param fileRef0
     */
    public void setRefFile0_FileName( String fileRef0 )
    {
        m_referenceDicomFile0Path = fileRef0;
    }
    
    /**
     * Sets the file reference of image reference dicom file 1
     * (ie a file system path).
     * @param fileRef1
     */
    public void setRefFile1_FileName( String fileRef1 )
    {
        m_referenceDicomFile1Path = fileRef1;
    }

    /**
     * Sets the file reference of image reference dicom file 2
     * (ie a file system path).
     * @param fileRef2
     */
    public void setRefFile2_FileName( String fileRef2 )
    {
        m_referenceDicomFile2Path = fileRef2;
    }

    /**
     * Returns the echo time of the dataset as integer.
     * @return 
     */
    public int getTE( )
    {
        return m_TE;
    }
    
    /**
     * Sets the acquisition frequency of the dataset.
     * @param acqFreq
     */
    public void setAcquisitionFrequency( float acqFreq )
    {
        m_acqFrequency = acqFreq;
    }
    
    /**
     * Sets the thickness of the excited single voxel spectroscopy volume. 
     * @param sVOIThickness 
     */
     public void set_sVOIThickness( double sVOIThickness )
     {
         m_sVOIThickness = sVOIThickness;
     }
     
    /**
     * Sets the thickness of the excited single voxel spectroscopy volume in 
     * the phase direction. 
     * @param sVoIPhaseFOV 
     */
     public void set_sVoIPhaseFOV( double sVoIPhaseFOV )
     {
        m_sVoIPhaseFOV = sVoIPhaseFOV;
     }
     
     /**
     * Sets the thickness of the excited single voxel spectroscopy volume in 
     * the readout direction. 
     * @param sVoIReadoutFOV 
     */
     public void set_sVoIReadoutFOV( double sVoIReadoutFOV )
     {
        m_sVoIReadoutFOV = sVoIReadoutFOV;
     }
    
    /**
     * Returns the thickness of the excited single voxel spectroscopy volume. 
     * @return 
     */
     public double get_sVOIThickness( )
     {
         return m_sVOIThickness;
     }
     
    /**
     * Returns the thickness of the excited single voxel spectroscopy volume in 
     * the phase direction. 
     * @return 
     */
     public double get_sVoIPhaseFOV( )
     {
        return m_sVoIPhaseFOV;
     }
     
     /**
     * Returns the thickness of the excited single voxel spectroscopy volume in 
     * the readout direction. 
     * @return 
     */
     public double get_sVoIReadoutFOV( )
     {
        return m_sVoIReadoutFOV;
     }
     
    /**
     * Returns the excited volume as String.
     * @return 
     */ 
    public String getExcitedVolumeAsString( )
    {
        NumberFormat format = new DecimalFormat( "###.#" );
        
        String output = "[" + format.format( m_sVOIThickness ) + "," 
                            + format.format( m_sVoIPhaseFOV )+ "," 
                            + format.format( m_sVoIReadoutFOV )+ "]";
        
        return output;
    }
    
    
    // Attributes
    // private float m_sampleFrequency = 1000.0f;
    private int m_numberOfSamplePoints = 0;  
    // Both channel ( alterating abs, disp ) dispersion and absorption channel
    // in time domain (TD)
    protected float[] m_bothChannelsTD;
    protected float[] m_absorptionChannelTD;
    protected float[] m_dispersionChannelTD;
    protected float[] m_xaxisTD;
    
    // Both channel ( alterating abs, disp ) dispersion and absorption channel
    // in frequency domain (FD)
    protected float[] m_bothChannelsFD;
    protected float[] m_absorptionChannelFD;
    protected float[] m_dispersionChannelFD;
    protected float[] m_xaxisFD;
    
    protected FileInfo m_dicomHeaderInfo;
    
    private Color m_curveColor = Color.black;

    private int m_curveDrawingStyle = Curve.NORMAL_LINE;

    private String m_curveLabel = "";

    // Contains Siemens specific header information
    private LinkedHashMap m_siemensHeaderInfo;
    
    // File pointers to files on which the single voxel spectroscopy was
    // planned
    private String m_referenceDicomFile0Path;
    private String m_referenceDicomFile1Path;
    private String m_referenceDicomFile2Path;
    
    private Spectroscopic3DVoxel m_voxelGeometry;
    
    private double m_sVOIThickness          = 0.00;
    private double m_sVoIPhaseFOV           = 0.00;
    private double m_sVoIReadoutFOV         = 0.00;
    
    private double m_dSagVoxelPosition      = 0.00;
    private double m_dCorVoxelPosition      = 0.00;
    private double m_dTraVoxelPosition      = 0.00;
    
    private double m_dSagVoxelOrientation   = 0.00;
    private double m_dCorVoxelOrientation   = 0.00;
    private double m_dTraVoxelOrientation   = 0.00;
    
    private double m_dInPlaneRotation;
    
    protected int    m_TE;
    protected double m_TR;
    
    // Value necessary to compute exact ppm values
    private long    m_resonanceFrequency        = 63000000L;
    private String  m_nucluesType               = "1H";
    
    // Value from Siemens specific header to compute sample frequency
    private float   m_sampleDwellTime         = 833;
    private float   m_acqFrequency            = 1200;
   
    
    public final static int TIME_DOMAIN     = 1;
    public final static int FREQ_DOMAIN     = 2;
}
